"""File Functions."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from daft.expressions import Expression
    from daft.io import IOConfig


def file(url: Expression, io_config: IOConfig | None = None) -> Expression:
    """Converts a string containing a file reference to a `daft.File` reference.

    Args:
        url (StringExpression): the url of the file
        io_config (IOConfig, default=None): The IO configuration to use.

    Returns:
        Expression (File Expression): An expression containing the file reference.

    """
    return url._eval_expressions("file", io_config=io_config)


def video_file(url: Expression, verify: bool = False, io_config: IOConfig | None = None) -> Expression:
    """Converts a string containing a file reference to a `daft.VideoFile` reference.

    Args:
        url (String Expression): the url of the file
        verify:
            If True, verify that the file exists and is a video file.
            If **ANY** files are not videos, this will produce an error.

        io_config (IOConfig, default=None): The IO configuration to use.

    Returns:
        Expression (File[Video] Expression): An expression containing the file reference.

    """
    return url._eval_expressions("video_file", verify=verify, io_config=io_config)


def audio_file(url: Expression, verify: bool = False, io_config: IOConfig | None = None) -> Expression:
    """Converts a string containing a file reference to a `daft.AudioFile` reference.

    Args:
        url (String Expression): the url of the file
        verify:
            If True, verify that the file exists and is a audio file.
            If **ANY** files are not audios, this will produce an error.

        io_config (IOConfig, default=None): The IO configuration to use.

    Returns:
        Expression (File[Audio] Expression): An expression containing the file reference.

    """
    return url._eval_expressions("audio_file", verify=verify, io_config=io_config)


def file_size(file: Expression) -> Expression:
    """Returns the size of the file in bytes.

    Args:
        file (File Expression): expression to evaluate.

    Returns:
        Expression (UInt64 Expression): expression containing the file size in bytes
    """
    return file._eval_expressions("file_size")


def guess_mime_type(bytes_expr: Expression) -> Expression:
    r"""Guess the MIME type of binary data by inspecting magic bytes.

    Detects common file formats including: PNG, JPEG, GIF, WEBP, PDF, ZIP,
    MP3, WAV, OGG, MP4, MPEG, and HTML. Returns None if the format cannot
    be determined.

    Args:
        bytes_expr: Binary expression containing raw bytes.

    Returns:
        Expression: String expression containing MIME type (e.g., "image/png") or None.

    Example:
        >>> import daft
        >>> from daft.functions import guess_mime_type
        >>> df = daft.from_pydict({"data": [b"\x89PNG\r\n\x1a\n", b"not a known format"]})
        >>> df = df.with_column("mime", guess_mime_type(df["data"]))
        >>> df.select("mime").show()
        ╭───────────╮
        │ mime      │
        │ ---       │
        │ String    │
        ╞═══════════╡
        │ image/png │
        ├╌╌╌╌╌╌╌╌╌╌╌┤
        │ None      │
        ╰───────────╯
        <BLANKLINE>
        (Showing first 2 of 2 rows)

    """
    return bytes_expr._eval_expressions("guess_mime_type")



from __future__ import annotations

import os
import tempfile
from contextlib import contextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, TypeAlias
from urllib.parse import urlparse

import av

from daft.daft import FileInfos, ImageMode
from daft.datatype import DataType
from daft.filesystem import _infer_filesystem, glob_path_with_stats
from daft.io import DataSource, DataSourceTask
from daft.recordbatch import MicroPartition
from daft.schema import Schema

if TYPE_CHECKING:
    from collections.abc import Generator, Iterator
    from fractions import Fraction

    from av.video import VideoFrame

    from daft.daft import IOConfig
    from daft.io.pushdowns import Pushdowns


if TYPE_CHECKING:
    from daft.dependencies import np

    _VideoFrameData: TypeAlias = np.typing.NDArray[Any]
else:
    _VideoFrameData: TypeAlias = Any


@dataclass
class _VideoFrame:
    """Represents a single video frame.

    Note:
        The field name 'data' is required due to a casting bug.
        See: https://github.com/Eventual-Inc/Daft/issues/4872
    """

    path: str
    frame_index: int
    frame_time: float
    frame_time_base: Fraction
    frame_pts: int
    frame_dts: int | None
    frame_duration: int | None
    is_key_frame: bool
    data: _VideoFrameData


@dataclass
class _VideoFramesSource(DataSource):
    """DataSource for streaming video files into rows of images.

    Note:
        This API requires [PyAV](https://github.com/PyAV-Org/PyAV), please do `pip install av`.

    Attributes:
        paths (list[str]): Path(s) to the video file(s).
        image_height (int): Height to which each frame will be resized.
        image_width (int): Width to which each frame will be resized.
        is_key_frame (bool|None): If True, only include key frames; if False, only non-key frames; if None, include all frames.
        io_config (IOConfig|None): Optional IOConfig.
    """

    paths: list[str]
    image_height: int
    image_width: int
    is_key_frame: bool | None = None
    io_config: IOConfig | None = None
    sample_interval_seconds: float | None = None

    @property
    def name(self) -> str:
        return "VideoFramesSource"

    @property
    def schema(self) -> Schema:
        return _schema(
            image_height=self.image_height,
            image_width=self.image_width,
        )

    def _list_file_infos(self) -> Generator[FileInfos]:
        for file_paths in self.paths:
            yield glob_path_with_stats(file_paths, file_format=None, io_config=self.io_config)

    def _list_file_paths(self) -> Generator[str]:
        all_youtube = all(_is_youtube_url(p) for p in self.paths)
        any_youtube = any(_is_youtube_url(p) for p in self.paths)
        if all_youtube:
            _assert_youtube_available()
            yield from self.paths
        elif any_youtube:
            raise ValueError("Either all or none of the paths must be YouTube URLs.")
        else:
            for file_infos in self._list_file_infos():
                yield from file_infos.file_paths

    def get_tasks(self, pushdowns: Pushdowns) -> Iterator[DataSourceTask]:
        for path in self._list_file_paths():
            yield _VideoFramesSourceTask(
                path=path,
                image_height=self.image_height,
                image_width=self.image_width,
                is_key_frame=self.is_key_frame,
                io_config=self.io_config,
                sample_interval_seconds=self.sample_interval_seconds,
            )


@dataclass
class _VideoFramesSourceTask(DataSourceTask):
    """DataSourceTask which yields micropartitions of images from a video file."""

    path: str
    image_height: int
    image_width: int
    is_key_frame: bool | None
    io_config: IOConfig | None
    sample_interval_seconds: float | None = None

    _max_partition_size = 10 * 1024 * 1024  # 10 MB

    @property
    def schema(self) -> Schema:
        return _schema(
            image_height=self.image_height,
            image_width=self.image_width,
        )

    def _list_frames(self, path: str, file: Any) -> Generator[_VideoFrame]:
        container = None
        try:
            container = av.open(file)

            # TODO support reading frames for multiple video streams
            stream = next((s for s in container.streams if s.type == "video"), None)
            if stream is None:
                container.close()
                raise RuntimeError(f"No video stream found in file: {path}")

            if self.is_key_frame:
                # https://pyav.org/docs/develop/cookbook/basics.html#saving-keyframes
                stream.codec_context.skip_frame = "NONKEY"

            sample_interval = self.sample_interval_seconds
            if sample_interval is not None and sample_interval <= 0:
                raise ValueError("sample_interval_seconds must be positive if provided")

            next_sample_time: float | None = 0.0 if sample_interval is not None else None
            # Small tolerance for floating point comparisons
            epsilon: float = 1e-9 if sample_interval is None else max(1e-9, sample_interval * 1e-6)

            frame_index: int = 0
            frame: VideoFrame
            while True:
                try:
                    frame = next(container.decode(stream))
                except av.EOFError:
                    break
                except StopIteration:
                    break

                frame_time = frame.time
                should_emit = True

                if sample_interval is not None:
                    if frame_time is None:
                        # Skip frames without timestamps when sampling is enabled.
                        should_emit = False
                    else:
                        assert next_sample_time is not None
                        if frame_time + epsilon < next_sample_time:
                            should_emit = False
                        else:
                            # Emit this frame and advance to the next target timestamp(s).
                            should_emit = True
                            while next_sample_time is not None and frame_time + epsilon >= next_sample_time:
                                next_sample_time += sample_interval

                if should_emit:
                    reformatted = frame.reformat(
                        width=self.image_width,
                        height=self.image_height,
                    )

                    yield _VideoFrame(
                        path=path,
                        frame_index=frame_index,
                        frame_time=frame_time,
                        frame_time_base=frame.time_base,
                        frame_pts=frame.pts,
                        frame_dts=frame.dts,
                        frame_duration=frame.duration,
                        is_key_frame=frame.key_frame,
                        data=reformatted.to_ndarray(format="rgb24"),
                    )

                frame_index += 1
        finally:
            if container:
                container.close()

    def _open(self) -> Any:
        if _is_youtube_url(self.path):
            return self._open_youtube_file()
        else:
            fp, fs, _ = _infer_filesystem(self.path, io_config=self.io_config)
            return fs.open_input_file(fp)

    @contextmanager
    def _open_youtube_file(self) -> Any:
        import yt_dlp

        def selector(ctx):  # type: ignore
            best_fmt = None
            best_fit = float("inf")  # lower is better

            for fmt in ctx["formats"]:
                if fmt.get("ext") != "mp4":
                    continue

                h, w = fmt.get("height"), fmt.get("width")
                if h is None or w is None:
                    continue

                fit = abs(h - self.image_height) + abs(w - self.image_width)
                if fit == 0:
                    yield fmt
                    return
                if fit < best_fit:
                    best_fmt = fmt
                    best_fit = fit

            yield best_fmt

        # Note:
        #   Streaming youtube downloads requires deeper work and was error-prone.
        #   The parsed youtube urls use m3u8 which requires decoding which ydl
        #   will handle for us. We cannot reliable pass a file-like HTTP response
        #   to PyAV; hence why this will download to a tempfile then open the file.

        temp_file = next(tempfile._get_candidate_names())  # type: ignore

        params = {
            "format": selector,
            "quiet": True,
            "outtmpl": temp_file,
            "no_warnings": True,
            "extract_flat": False,
            "no_check_certificate": True,
            "ignoreerrors": False,
            "consoletitle": False,
            "noprogress": True,
        }

        try:
            with yt_dlp.YoutubeDL(params) as ydl:
                ydl.download([self.path])
            yield open(temp_file, mode="rb")
        finally:
            if os.path.exists(temp_file):
                os.remove(temp_file)

    def get_micro_partitions(self) -> Iterator[MicroPartition]:
        with self._open() as file:
            buffer = _VideoFramesBuffer(
                image_height=self.image_height,
                image_width=self.image_width,
            )
            for frame in self._list_frames(self.path, file):
                buffer.append(frame)
                # yield when full
                if buffer.size() >= self._max_partition_size:
                    yield buffer.to_micropartition()
                    buffer.clear()
            # yield if non-empty
            if buffer and buffer.size() > 0:
                yield buffer.to_micropartition()


class _VideoFramesBuffer:
    """A micropartition buffer/builder for video frames.

    Note:
        This enables decoupling the video source from a particular
        library, making it possible for the VideoSource to leverage
        an Iterable[_VideoFrame[T]] at some later time. How the iterator
        is implemented e.g. open-cv vs. PyAV or other library is not
        important, just that we have an Iterable of frames which this
        builder and the source and stream as appropriately sized partitions.
    """

    image_height: int
    image_width: int

    _arr_path: list[str]
    _arr_frame_index: list[int]
    _arr_frame_time: list[float]
    _arr_frame_time_base: list[str]
    _arr_frame_pts: list[int]
    _arr_frame_dts: list[int | None]
    _arr_frame_duration: list[int | None]
    _arr_is_key_frame: list[bool]
    _arr_data: list[_VideoFrameData]
    _size_in_bytes: int
    _size_of_metadata = 64

    def __init__(self, image_height: int, image_width: int):
        self.image_height = image_height
        self.image_width = image_width
        self.clear()

    def clear(self) -> None:
        self._arr_path = []
        self._arr_frame_index = []
        self._arr_frame_time = []
        self._arr_frame_time_base = []
        self._arr_frame_pts = []
        self._arr_frame_dts = []
        self._arr_frame_duration = []
        self._arr_is_key_frame = []
        self._arr_data = []
        self._size_in_bytes = 0

    def size(self) -> int:
        """Returns the current size of the partition in bytes."""
        return self._size_in_bytes

    def append(self, frame: _VideoFrame) -> None:
        """Appends the frame to this partition builder.

        Note:
            We encode the time_base as a fraction string.
            See: https://github.com/Eventual-Inc/Daft/issues/4971
        """
        self._arr_path.append(frame.path)
        self._arr_frame_index.append(frame.frame_index)
        self._arr_frame_time.append(frame.frame_time)
        self._arr_frame_time_base.append(str(frame.frame_time_base))
        self._arr_frame_pts.append(frame.frame_pts)
        self._arr_frame_dts.append(frame.frame_dts)
        self._arr_frame_duration.append(frame.frame_duration)
        self._arr_is_key_frame.append(frame.is_key_frame)
        self._arr_data.append(frame.data)
        self._size_in_bytes += frame.data.nbytes + self._size_of_metadata

    def to_micropartition(self) -> MicroPartition:
        """Returns a MicroPartition for this builder."""
        return MicroPartition.from_pydict(
            {
                "path": self._arr_path,
                "frame_index": self._arr_frame_index,
                "frame_time": self._arr_frame_time,
                "frame_time_base": self._arr_frame_time_base,
                "frame_pts": self._arr_frame_pts,
                "frame_dts": self._arr_frame_dts,
                "frame_duration": self._arr_frame_duration,
                "is_key_frame": self._arr_is_key_frame,
                "data": self._arr_data,
            }
        )


def _schema(image_height: int, image_width: int) -> Schema:
    """Returns the common schema which is needed in several places."""
    return Schema.from_pydict(
        {
            "path": DataType.string(),
            "frame_index": DataType.int64(),
            "frame_time": DataType.float64(),
            "frame_time_base": DataType.string(),
            "frame_pts": DataType.int64(),
            "frame_dts": DataType.int64(),
            "frame_duration": DataType.int64(),
            "is_key_frame": DataType.bool(),
            "data": DataType.image(
                height=image_height,
                width=image_width,
                mode=ImageMode.RGB,  # make configurable at later time
            ),
        }
    )


def _is_youtube_url(url: str) -> bool:
    hostname = urlparse(url).hostname
    if hostname is None:
        return False
    return hostname in {"www.youtube.com", "youtube.com", "youtu.be"}


def _assert_youtube_available() -> None:
    try:
        import yt_dlp  # noqa: F401
    except ImportError as e:
        raise ImportError(
            "YouTube video support requires the 'yt-dlp' package. Please install it with `pip install yt-dlp`."
        ) from e



"""URL Functions."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from daft import context
from daft.expressions import Expression
from daft.runners import get_or_create_runner

if TYPE_CHECKING:
    from daft.daft import IOConfig


def _should_use_multithreading_tokio_runtime() -> bool:
    """Whether or not our expression should use the multithreaded tokio runtime under the hood, or a singlethreaded one.

    This matters because for distributed workloads, each process has its own tokio I/O runtime. if each distributed process
    is multithreaded (by default we spin up `N_CPU` threads) then we will be running `(N_CPU * N_PROC)` number of threads, and
    opening `(N_CPU * N_PROC * max_connections)` number of connections. This is too large for big machines with many CPU cores.

    Hence for Ray we default to doing the singlethreaded runtime. This means that we will have a limit of
    `(singlethreaded=1 * N_PROC * max_connections)` number of open connections per machine, which works out to be reasonable at ~2-4k connections.

    For local execution, we run in a single process which means that it all shares the same tokio I/O runtime and connection pool.
    Thus we just have `(multithreaded=N_CPU * max_connections)` number of open connections, which is usually reasonable as well.
    """
    using_ray_runner = get_or_create_runner().name == "ray"
    return not using_ray_runner


def _override_io_config_max_connections(max_connections: int, io_config: IOConfig | None) -> IOConfig:
    """Use a user-provided `max_connections` argument to override the value in S3Config.

    This is because our Rust code under the hood actually does `min(S3Config's max_connections, url_download's max_connections)` to
    determine how many connections to allow per-thread. Thus we need to override the io_config here to ensure that the user's max_connections
    is correctly applied in our Rust code.
    """
    io_config = context.get_context().daft_planning_config.default_io_config if io_config is None else io_config
    io_config = io_config.replace(s3=io_config.s3.replace(max_connections=max_connections))
    return io_config


def download(
    expr: Expression,
    max_connections: int = 32,
    on_error: Literal["raise", "null"] = "raise",
    io_config: IOConfig | None = None,
) -> Expression:
    """Treats each string as a URL, and downloads the bytes contents as a bytes column.

    Args:
        expr: The expression to download.
        max_connections: The maximum number of connections to use per thread to use for downloading URLs. Defaults to 32.
        on_error: Behavior when a URL download error is encountered - "raise" to raise the error immediately or "null" to log
            the error but fallback to a Null value. Defaults to "raise".
        io_config: IOConfig to use when accessing remote storage. Note that the S3Config's `max_connections` parameter will be overridden
            with `max_connections` that is passed in as a kwarg.

    Returns:
        Expression: a Binary expression which is the bytes contents of the URL, or None if an error occurred during download

    Note:
        If you are observing excessive S3 issues (such as timeouts, DNS errors or slowdown errors) during URL downloads,
        you may wish to reduce the value of ``max_connections`` (defaults to 32) to reduce the amount of load you are placing
        on your S3 servers.

        Alternatively, if you are running on machines with lower number of cores but very high network bandwidth, you can increase
        ``max_connections`` to get higher throughput with additional parallelism
    """
    multi_thread = _should_use_multithreading_tokio_runtime()
    io_config = _override_io_config_max_connections(max_connections, io_config)

    if io_config.unity.endpoint is None:
        try:
            from daft.catalog.__unity import UnityCatalog
        except ImportError:
            pass
        else:
            from daft.session import current_catalog

            catalog = current_catalog()
            if isinstance(catalog, UnityCatalog):
                unity_catalog = catalog._inner
                io_config = io_config.replace(unity=unity_catalog.to_io_config().unity)

    return Expression._call_builtin_scalar_fn(
        "url_download",
        expr,
        multi_thread=multi_thread,
        on_error=on_error,
        max_connections=max_connections,
        io_config=io_config,
    )


def upload(
    expr: Expression,
    location: str | Expression,
    max_connections: int = 32,
    on_error: Literal["raise", "null"] = "raise",
    io_config: IOConfig | None = None,
) -> Expression:
    """Uploads a column of binary data to the provided location(s) (also supports S3, local etc).

    Files will be written into the location (folder(s)) with a generated UUID filename, and the result
    will be returned as a column of string paths that is compatible with the ``download()`` Expression.

    Args:
        expr: The expression to upload.
        location: a folder location or column of folder locations to upload data into
        max_connections: The maximum number of connections to use per thread to use for uploading data. Defaults to 32.
        on_error: Behavior when a URL upload error is encountered - "raise" to raise the error immediately or "null" to log
            the error but fallback to a Null value. Defaults to "raise".
        io_config: IOConfig to use when uploading data

    Returns:
        Expression: a String expression containing the written filepath

    Examples:
        >>> from daft.functions import upload
        >>>
        >>> upload(df["data"], "s3://my-bucket/my-folder")  # doctest: +SKIP

        Upload to row-specific URLs

        >>> upload(df["data"], df["paths"])  # doctest: +SKIP

    """
    multi_thread = _should_use_multithreading_tokio_runtime()
    # If the user specifies a single location via a string, we should upload to a single folder. Otherwise,
    # if the user gave an expression, we assume that each row has a specific url to upload to.
    # Consider moving the check for is_single_folder to a lower IR.
    is_single_folder = isinstance(location, str)
    io_config = _override_io_config_max_connections(max_connections, io_config)

    return Expression._call_builtin_scalar_fn(
        "url_upload",
        expr,
        location,
        max_connections=max_connections,
        on_error=on_error,
        multi_thread=multi_thread,
        is_single_folder=is_single_folder,
        io_config=io_config,
    )


def parse_url(expr: Expression) -> Expression:
    """Parse string URLs and extract URL components.

    Returns:
        Expression: a Struct expression containing the parsed URL components:
            - scheme (str): The URL scheme (e.g., "https", "http")
            - username (str): The username, if present
            - password (str): The password, if present
            - host (str): The hostname or IP address
            - port (int): The port number, if specified
            - path (str): The path component
            - query (str): The query string, if present
            - fragment (str): The fragment/anchor, if present

    Examples:
        >>> import daft
        >>> from daft.functions import parse_url
        >>>
        >>> df = daft.from_pydict(
        ...     {"urls": ["https://user:pass@example.com:8080/path?query=value#fragment", "http://localhost/api"]}
        ... )
        >>> # Parse URLs and expand all components
        >>> df.select(parse_url(df["urls"]).unnest()).collect()  # doctest: +SKIP

    Note:
        Invalid URLs will result in null values for all components.
        The parsed result is automatically aliased to 'urls' to enable easy struct field expansion.
    """
    return Expression._call_builtin_scalar_fn("url_parse", expr)



"""Bitwise Functions."""

from __future__ import annotations

from daft.expressions import Expression


def bitwise_and(left: Expression, right: Expression) -> Expression:
    """Bitwise AND of two integer expressions."""
    left = Expression._to_expression(left)
    right = Expression._to_expression(right)
    return Expression._from_pyexpr(left._expr & right._expr)


def bitwise_or(left: Expression, right: Expression) -> Expression:
    """Bitwise OR of two integer expressions."""
    left = Expression._to_expression(left)
    right = Expression._to_expression(right)
    return Expression._from_pyexpr(left._expr | right._expr)


def bitwise_xor(left: Expression, right: Expression) -> Expression:
    """Bitwise XOR of two integer expressions."""
    left = Expression._to_expression(left)
    right = Expression._to_expression(right)
    return Expression._from_pyexpr(left._expr ^ right._expr)


def shift_left(expr: Expression, num_bits: Expression) -> Expression:
    """Shifts the bits of an integer expression to the left (``expr << num_bits``).

    Args:
        expr: The expression to shift.
        num_bits: The number of bits to shift the expression to the left
    """
    expr = Expression._to_expression(expr)
    num_bits = Expression._to_expression(num_bits)
    return Expression._from_pyexpr(expr._expr << num_bits._expr)


def shift_right(expr: Expression, num_bits: Expression) -> Expression:
    """Shifts the bits of an integer expression to the right (``expr >> num_bits``).

    Args:
        expr: The expression to shift.
        num_bits: The number of bits to shift the expression to the right

    Note:
        For unsigned integers, this expression perform a logical right shift.
        For signed integers, this expression perform an arithmetic right shift.
    """
    expr = Expression._to_expression(expr)
    num_bits = Expression._to_expression(num_bits)
    return Expression._from_pyexpr(expr._expr >> num_bits._expr)



from __future__ import annotations

import contextlib
import logging
import threading
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ClassVar

from daft.daft import IOConfig, PyDaftContext, PyDaftExecutionConfig, PyDaftPlanningConfig, PyQueryResult
from daft.daft import get_context as _get_context

if TYPE_CHECKING:
    from collections.abc import Generator

    from daft.daft import PyQueryMetadata
    from daft.runners.partitioning import PartitionT
    from daft.subscribers import Subscriber

logger = logging.getLogger(__name__)


@dataclass
class DaftContext:
    """Global context for the current Daft execution environment."""

    _ctx: PyDaftContext

    _lock: ClassVar[threading.Lock] = threading.Lock()

    @staticmethod
    def _from_native(ctx: PyDaftContext) -> DaftContext:
        return DaftContext(ctx=ctx)

    def __init__(self, ctx: PyDaftContext | None = None):
        if ctx is not None:
            self._ctx = ctx
        else:
            self._ctx = PyDaftContext()

    @property
    def daft_execution_config(self) -> PyDaftExecutionConfig:
        return self._ctx._daft_execution_config

    @property
    def daft_planning_config(self) -> PyDaftPlanningConfig:
        return self._ctx._daft_planning_config

    def attach_subscriber(self, alias: str, subscriber: Subscriber) -> None:
        """Attaches a subscriber to this context.

        Subscribers listen to events emitted during runtime, particularly during query execution.
        See the Subscriber class for more details.

        Args:
            alias (str): alias for the subscriber
            subscriber (Subscriber): subscriber instance
        """
        self._ctx.attach_subscriber(alias, subscriber)

    def detach_subscriber(self, alias: str) -> None:
        """Detaches a subscriber from this context.

        Args:
            alias (str): alias for the subscriber
        """
        self._ctx.detach_subscriber(alias)

    def _notify_query_start(self, query_id: str, metadata: PyQueryMetadata) -> None:
        self._ctx.notify_query_start(query_id, metadata)

    def _notify_query_end(self, query_id: str, query_result: PyQueryResult) -> None:
        """Notifies the query end to the subscribers. Exceptions from subscribers are logged and ignored."""
        self._ctx.notify_query_end(query_id, query_result)

    def _notify_optimization_start(self, query_id: str) -> None:
        self._ctx.notify_optimization_start(query_id)

    def _notify_optimization_end(self, query_id: str, optimized_plan: str) -> None:
        self._ctx.notify_optimization_end(query_id, optimized_plan)

    def _notify_exec_start(self, query_id: str, physical_plan: str) -> None:
        self._ctx.notify_exec_start(query_id, physical_plan)

    def _notify_exec_end(self, query_id: str) -> None:
        self._ctx.notify_exec_end(query_id)

    def _notify_exec_operator_start(self, query_id: str, node_id: int) -> None:
        self._ctx.notify_exec_operator_start(query_id, node_id)

    def _notify_exec_operator_end(self, query_id: str, node_id: int) -> None:
        self._ctx.notify_exec_operator_end(query_id, node_id)

    def _notify_exec_emit_stats(self, query_id: str, node_id: int, stats: dict[str, int]) -> None:
        self._ctx.notify_exec_emit_stats(query_id, node_id, stats)

    def _notify_result_out(self, query_id: str, result: PartitionT) -> None:
        from daft.recordbatch.micropartition import MicroPartition

        if not isinstance(result, MicroPartition):
            raise ValueError("Query Managers only support the Native Runner for now")
        self._ctx.notify_result_out(query_id, result._micropartition)


def get_context() -> DaftContext:
    """Returns the global singleton daft context."""
    return DaftContext(_get_context())


@contextlib.contextmanager
def planning_config_ctx(**kwargs: Any) -> Generator[None, None, None]:
    """Context manager that wraps set_planning_config to reset the config to its original setting afternwards."""
    original_config = get_context().daft_planning_config
    try:
        set_planning_config(**kwargs)
        yield
    finally:
        set_planning_config(config=original_config)


def set_planning_config(
    config: PyDaftPlanningConfig | None = None,
    default_io_config: IOConfig | None = None,
    enable_strict_filter_pushdown: bool | None = None,
) -> DaftContext:
    """Globally sets various configuration parameters which control Daft plan construction behavior.

    These configuration values are used when a Dataframe is being constructed (e.g. calls to create a Dataframe, or to build on an existing Dataframe).

    Args:
        config: A PyDaftPlanningConfig object to set the config to, before applying other kwargs. Defaults to None which indicates
            that the old (current) config should be used.
        default_io_config: A default IOConfig to use in the absence of one being explicitly passed into any Expression (e.g. `.download()`)
            or Dataframe operation (e.g. `daft.read_parquet()`).
    """
    # Replace values in the DaftPlanningConfig with user-specified overrides
    ctx = get_context()
    with ctx._lock:
        old_daft_planning_config = ctx._ctx._daft_planning_config if config is None else config
        new_daft_planning_config = old_daft_planning_config.with_config_values(
            default_io_config=default_io_config, enable_strict_filter_pushdown=enable_strict_filter_pushdown
        )

        ctx._ctx._daft_planning_config = new_daft_planning_config
        return ctx


@contextlib.contextmanager
def execution_config_ctx(**kwargs: Any) -> Generator[None, None, None]:
    """Context manager that wraps set_execution_config to reset the config to its original setting afternwards."""
    original_config = get_context()._ctx._daft_execution_config
    try:
        set_execution_config(**kwargs)
        yield
    finally:
        set_execution_config(config=original_config)


def set_execution_config(
    config: PyDaftExecutionConfig | None = None,
    enable_scan_task_split_and_merge: bool | None = None,
    scan_tasks_min_size_bytes: int | None = None,
    scan_tasks_max_size_bytes: int | None = None,
    max_sources_per_scan_task: int | None = None,
    broadcast_join_size_bytes_threshold: int | None = None,
    parquet_split_row_groups_max_files: int | None = None,
    hash_join_partition_size_leniency: float | None = None,
    sample_size_for_sort: int | None = None,
    num_preview_rows: int | None = None,
    parquet_target_filesize: int | None = None,
    parquet_target_row_group_size: int | None = None,
    parquet_inflation_factor: float | None = None,
    csv_target_filesize: int | None = None,
    csv_inflation_factor: float | None = None,
    json_inflation_factor: float | None = None,
    text_inflation_factor: float | None = None,
    shuffle_aggregation_default_partitions: int | None = None,
    partial_aggregation_threshold: int | None = None,
    high_cardinality_aggregation_threshold: float | None = None,
    read_sql_partition_size_bytes: int | None = None,
    default_morsel_size: int | None = None,
    shuffle_algorithm: str | None = None,
    pre_shuffle_merge_threshold: int | None = None,
    scantask_max_parallel: int | None = None,
    native_parquet_writer: bool | None = None,
    min_cpu_per_task: float | None = None,
    actor_udf_ready_timeout: int | None = None,
    maintain_order: bool | None = None,
    enable_dynamic_batching: bool | None = None,
    dynamic_batching_strategy: str | None = None,
    flight_shuffle_dirs: list[str] | None = None,
    enable_multi_glob_path_tasks: bool | None = None,
) -> DaftContext:
    """Globally sets various configuration parameters which control various aspects of Daft execution.

    These configuration values
    are used when a Dataframe is executed (e.g. calls to `DataFrame.write_*`, [DataFrame.collect()](https://docs.daft.ai/en/stable/api/dataframe/#daft.DataFrame.collect) or [DataFrame.show()](https://docs.daft.ai/en/stable/api/dataframe/#daft.DataFrame.select)).

    Args:
        config: A PyDaftExecutionConfig object to set the config to, before applying other kwargs. Defaults to None which indicates
            that the old (current) config should be used.
        enable_scan_task_split_and_merge: Whether to enable scan task split and merge. Defaults to False.
        scan_tasks_min_size_bytes: Minimum size of scan tasks in bytes. Defaults to 96MB.
        scan_tasks_max_size_bytes: Maximum size of scan tasks in bytes. Defaults to 384MB.
        max_sources_per_scan_task: Maximum number of sources per scan task. Defaults to 10.
        parquet_split_row_groups_max_files: Maximum number of files to read in which the row group splitting should happen. (Defaults to 10)
        broadcast_join_size_bytes_threshold: If one side of a join is smaller than this threshold, a broadcast join will be used.
            Default is 10 MiB.
        hash_join_partition_size_leniency: If the left side of a hash join is already correctly partitioned and the right side isn't,
            and the ratio between the left and right size is at least this value, then the right side is repartitioned to have an equal
            number of partitions as the left. Defaults to 0.5.
        sample_size_for_sort: number of elements to sample from each partition when running sort,
            Default is 20.
        num_preview_rows: number of rows to when showing a dataframe preview,
            Default is 8.
        parquet_target_filesize: Target File Size when writing out Parquet Files. Defaults to 512MB
        parquet_target_row_group_size: Target Row Group Size when writing out Parquet Files. Defaults to 128MB
        parquet_inflation_factor: Inflation Factor of parquet files (In-Memory-Size / File-Size) ratio. Defaults to 3.0
        csv_target_filesize: Target File Size when writing out CSV Files. Defaults to 512MB
        csv_inflation_factor: Inflation Factor of CSV files (In-Memory-Size / File-Size) ratio. Defaults to 0.5
        json_inflation_factor: Inflation Factor of JSON files (In-Memory-Size / File-Size) ratio. Defaults to 0.25
        text_inflation_factor: Inflation Factor of Text files (In-Memory-Size / File-Size) ratio. Defaults to 1.0
        shuffle_aggregation_default_partitions: Maximum number of partitions to create when performing aggregations on the Ray Runner. Defaults to 200, unless the number of input partitions is less than 200.
        partial_aggregation_threshold: Threshold for performing partial aggregations on the Native Runner. Defaults to 10000 rows.
        high_cardinality_aggregation_threshold: Threshold selectivity for performing high cardinality aggregations on the Native Runner. Defaults to 0.8.
        read_sql_partition_size_bytes: Target size of partition when reading from SQL databases. Defaults to 512MB
        default_morsel_size: Default size of morsels used for the new local executor. Defaults to 131072 rows.
        shuffle_algorithm: The shuffle algorithm to use. Defaults to "auto", which will let Daft determine the algorithm. Options are "map_reduce", "pre_shuffle_merge", and "flight_shuffle".
        pre_shuffle_merge_threshold: Memory threshold in bytes for pre-shuffle merge. Defaults to 1GB
        scantask_max_parallel: Set the max parallelism for running scan tasks simultaneously. Currently, this only works for Native Runner. If set to 0, all available CPUs will be used. Defaults to 8.
        native_parquet_writer: Whether to use the native parquet writer vs the pyarrow parquet writer. Defaults to `True`.
        min_cpu_per_task: Minimum CPU per task in the Ray runner. Defaults to 0.5.
        actor_udf_ready_timeout: Timeout for UDF actors to be ready. Defaults to 120 seconds.
        maintain_order: Whether to maintain order during execution. Defaults to True. Some blocking sink operators (e.g. write_parquet) won't respect this flag and will always keep maintain_order as false, and propagate to child operators. It's useful to set this to False for running df.collect() when no ordering is required.
        enable_dynamic_batching: Whether to enable dynamic batching. Defaults to False.
        dynamic_batching_strategy: The strategy to use for dynamic batching. Defaults to 'auto'.
        flight_shuffle_dirs: Directories to use for flight shuffle. Defaults to ["/tmp"]. Must not be empty.
        enable_multi_glob_path_tasks: Whether to create multiple glob path tasks in Ray Runner to achieve parallel glob. Defaults to False.
    """
    # Replace values in the DaftExecutionConfig with user-specified overrides
    ctx = get_context()
    with ctx._lock:
        old_daft_execution_config = ctx._ctx._daft_execution_config if config is None else config

        new_daft_execution_config = old_daft_execution_config.with_config_values(
            enable_scan_task_split_and_merge=enable_scan_task_split_and_merge,
            scan_tasks_min_size_bytes=scan_tasks_min_size_bytes,
            scan_tasks_max_size_bytes=scan_tasks_max_size_bytes,
            max_sources_per_scan_task=max_sources_per_scan_task,
            parquet_split_row_groups_max_files=parquet_split_row_groups_max_files,
            broadcast_join_size_bytes_threshold=broadcast_join_size_bytes_threshold,
            hash_join_partition_size_leniency=hash_join_partition_size_leniency,
            sample_size_for_sort=sample_size_for_sort,
            num_preview_rows=num_preview_rows,
            parquet_target_filesize=parquet_target_filesize,
            parquet_target_row_group_size=parquet_target_row_group_size,
            parquet_inflation_factor=parquet_inflation_factor,
            csv_target_filesize=csv_target_filesize,
            csv_inflation_factor=csv_inflation_factor,
            json_inflation_factor=json_inflation_factor,
            text_inflation_factor=text_inflation_factor,
            shuffle_aggregation_default_partitions=shuffle_aggregation_default_partitions,
            partial_aggregation_threshold=partial_aggregation_threshold,
            high_cardinality_aggregation_threshold=high_cardinality_aggregation_threshold,
            read_sql_partition_size_bytes=read_sql_partition_size_bytes,
            default_morsel_size=default_morsel_size,
            shuffle_algorithm=shuffle_algorithm,
            pre_shuffle_merge_threshold=pre_shuffle_merge_threshold,
            scantask_max_parallel=scantask_max_parallel,
            native_parquet_writer=native_parquet_writer,
            min_cpu_per_task=min_cpu_per_task,
            actor_udf_ready_timeout=actor_udf_ready_timeout,
            maintain_order=maintain_order,
            enable_dynamic_batching=enable_dynamic_batching,
            dynamic_batching_strategy=dynamic_batching_strategy,
            flight_shuffle_dirs=flight_shuffle_dirs,
            enable_multi_glob_path_tasks=enable_multi_glob_path_tasks,
        )

        ctx._ctx._daft_execution_config = new_daft_execution_config
        return ctx



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from lance.optimize import Compaction, CompactionMetrics, CompactionOptions, CompactionTask, RewriteResult

if TYPE_CHECKING:
    import lance
import daft

logger = logging.getLogger(__name__)


class CompactionTaskUDF:
    """UDF to execute a batch of Lance CompactionTasks on remote workers and return execution result dictionaries."""

    def __init__(
        self,
        lance_ds: lance.LanceDataset,
    ) -> None:
        self.lance_ds = lance_ds

    def __call__(self, task: CompactionTask) -> RewriteResult:
        rewrite = task.execute(self.lance_ds)
        return rewrite


def compact_files_internal(
    lance_ds: lance.LanceDataset,
    *,
    compaction_options: dict[str, Any] | None = None,
    partition_num: int | None = None,
    concurrency: int | None = None,
) -> CompactionMetrics | None:
    """Execute Lance file compaction in distributed environment using Daft UDF style."""
    logger.info("Starting UDF-style distributed compaction")
    plan = Compaction.plan(
        lance_ds,
        CompactionOptions(
            **(compaction_options or {}),
        ),
    )
    num_tasks = plan.num_tasks()
    logger.info("Compaction plan created with %d tasks", num_tasks)

    if num_tasks == 0:
        logger.info("No compaction tasks needed")
        return None

    effective_partition_num = partition_num or 1
    effective_partition_num = min(num_tasks, effective_partition_num)
    assert effective_partition_num > 0
    if effective_partition_num == 1:
        df = daft.from_pydict({"task": plan.tasks})
    else:
        df = daft.from_pydict({"task": plan.tasks}).repartition(effective_partition_num)

    WrappedRunner = daft.cls(
        CompactionTaskUDF,
        max_concurrency=concurrency,
    )
    df = df.select(WrappedRunner(lance_ds)(df["task"]).alias("rewrite"))
    results = df.to_pandas()

    metrics = Compaction.commit(lance_ds, results["rewrite"].to_list())
    logger.info("Compaction completed successfully. Metrics: %s", metrics)
    return metrics



from __future__ import annotations

from typing import TYPE_CHECKING

from daft.context import get_context

if TYPE_CHECKING:
    from daft.dataframe.dataframe import DataFrame
    from daft.daft import IOConfig

__all__ = [
    # TODO: additional video support
    # "read_audio_frames",
    # "read_audio_streams",
    # "read_audio_streams_metadata",
    # "read_subtitle_frames",
    # "read_subtitle_streams",
    # "read_subtitle_streams_metadata",
    "read_video_frames",
    # "read_video_streams",
    # "read_video_streams_metadata",
]


def read_video_frames(
    path: str | list[str],
    image_height: int,
    image_width: int,
    is_key_frame: bool | None = None,
    *,
    sample_interval_seconds: float | None = None,
    io_config: IOConfig | None = None,
) -> DataFrame:
    """Creates a DataFrame by reading the frames of one or more video files.

    This produces a DataFrame with the following fields:
        * path (string): path to the video file that produced this frame.
        * frame_index (int): frame index in the video.
        * frame_time (float): frame time in fractional seconds as a floating point.
        * frame_time_base (str): fractional unit of seconds in which timestamps are expressed.
        * frame_pts (int): frame presentation timestamp in time_base units.
        * frame_dts (int): frame decoding timestamp in time_base units.
        * frame_duration (int): frame duration in time_base units.
        * is_key_frame (bool): true iff this is a key frame.

    Warning:
        This requires PyAV which can be installed with `pip install av`.

    Note:
        This function will stream the frames from all videos as a DataFrame of images.
        If you wish to load an entire video into a single row, this can be done with
        read_glob_path and url.download.

    Args:
        path (str|list[str]): Path(s) to the video file(s) which allows wildcards.
        image_height (int): Height to which each frame will be resized.
        image_width (int): Width to which each frame will be resized.
        is_key_frame (bool|None): If True, only include key frames; if False, only non-key frames; if None, include all frames.
        sample_interval_seconds (float|None): If provided and > 0, sample frames at approximately this time interval in seconds based on ``frame_time``.
            The algorithm selects the first frame whose timestamp is >= target time (0, interval, 2*interval, ...).
            This is an approximate sampling strategy; actual sampling times depend on the video's frame timestamps.
            Frames without valid timestamps (frame_time=None) are skipped.
        io_config (IOConfig|None): Optional IOConfig.

    Returns:
        DataFrame: dataframe of images.

    Examples:
        >>> df = daft.read_video_frames("/path/to/file.mp4", image_height=480, image_width=640)
        >>> df = daft.read_video_frames("/path/to/directory", image_height=480, image_width=640)
        >>> df = daft.read_video_frames("/path/to/files-*.mp4", image_height=480, image_width=640)
        >>> df = daft.read_video_frames("s3://path/to/files-*.mp4", image_height=480, image_width=640)

        Sample approximately one frame per second:
        >>> df = daft.read_video_frames(
        ...     "/path/to/file.mp4", image_height=480, image_width=640, sample_interval_seconds=1.0
        ... )

        Sample approximately one frame every 5 seconds:
        >>> df = daft.read_video_frames(
        ...     "/path/to/file.mp4", image_height=480, image_width=640, sample_interval_seconds=5.0
        ... )

        Combine with key frame filtering:
        >>> df = daft.read_video_frames(
        ...     "/path/to/file.mp4", image_height=480, image_width=640, is_key_frame=True, sample_interval_seconds=1.0
        ... )
    """
    try:
        from daft.io.av._read_video_frames import _VideoFramesSource
    except ImportError as e:
        raise ImportError("read_video_frames requires PyAV. Please install it with `pip install av`.") from e

    io_config = get_context().daft_planning_config.default_io_config if io_config is None else io_config
    return _VideoFramesSource(
        paths=[path] if isinstance(path, str) else path,
        image_height=image_height,
        image_width=image_width,
        is_key_frame=is_key_frame,
        io_config=io_config,
        sample_interval_seconds=sample_interval_seconds,
    ).read()



"""WARNING! These APIs are internal; please use Table.from_postgres()."""

from __future__ import annotations

import json
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any

import psycopg
from pgvector.psycopg import register_vector

from daft.catalog import Catalog, Identifier, NotFoundError, Properties, Schema, Table
from daft.datatype import DataType
from daft.io._sql import read_sql
from daft.logical.schema import Field


@contextmanager
def postgres_connection(connection_string: str, extensions: list[str] | None) -> psycopg.Connection.connect:
    """Context manager that provides a PostgreSQL connection with specified extensions setup.

    Args:
        connection_string: PostgreSQL connection string
        extensions: List of extension names to create if they don't exist and are available.
                   For each extension, availability is checked in pg_available_extensions before
                   attempting "CREATE EXTENSION IF NOT EXISTS <extension>".
    """
    with psycopg.connect(connection_string) as conn:
        if extensions:
            with conn.cursor() as cur:
                for extension in extensions:
                    # Check if extension is available before attempting to create it
                    cur.execute(
                        psycopg.sql.SQL("SELECT EXISTS(SELECT 1 FROM pg_available_extensions WHERE name = {})").format(
                            psycopg.sql.Literal(extension)
                        )
                    )
                    result = cur.fetchone()
                    is_available = result[0] if result else False

                    if is_available:
                        cur.execute(
                            psycopg.sql.SQL("CREATE EXTENSION IF NOT EXISTS {}").format(
                                psycopg.sql.Identifier(extension)
                            )
                        )

                # Register pgvector type if it was successfully created
                if "vector" in extensions:
                    cur.execute(psycopg.sql.SQL("SELECT EXISTS(SELECT 1 FROM pg_extension WHERE extname = 'vector')"))
                    result = cur.fetchone()
                    vector_installed = result[0] if result else False

                    if vector_installed:
                        register_vector(conn)
        yield conn


if TYPE_CHECKING:
    from daft.dataframe import DataFrame
    from daft.io.partitioning import PartitionField


def _daft_dtype_to_postgres_type(dtype: DataType, set_dimensions: bool = True) -> str:
    """Convert a Daft DataType to a PostgreSQL type string."""
    if dtype.is_int8():
        return "smallint"  # PostgreSQL doesn't have TINYINT, use SMALLINT
    elif dtype.is_int16():
        return "smallint"
    elif dtype.is_int32():
        return "integer"
    elif dtype.is_int64():
        return "bigint"
    elif dtype.is_uint8():
        return "smallint"  # PostgreSQL doesn't have unsigned types
    elif dtype.is_uint16():
        return "integer"  # PostgreSQL doesn't have unsigned types
    elif dtype.is_uint32():
        return "bigint"  # PostgreSQL doesn't have unsigned types
    elif dtype.is_uint64():
        return "bigint"  # PostgreSQL doesn't have unsigned types, may overflow
    elif dtype.is_float32():
        return "real"
    elif dtype.is_float64():
        return "double precision"
    elif dtype.is_boolean():
        return "boolean"
    elif dtype.is_string():
        return "text"
    elif dtype.is_binary() or dtype.is_fixed_size_binary():
        return "bytea"
    elif dtype.is_date():
        return "date"
    elif dtype.is_timestamp():
        # PostgreSQL timestamps support timezone
        try:
            timezone = dtype.timezone
            if timezone:
                return "timestamptz"
        except AttributeError:
            pass
        return "timestamp"
    elif dtype.is_time():
        return "time"
    # TODO(desmond): Daft is currently unable to read INTERVAL types via read_sql()
    # elif dtype.is_duration():
    #     return "interval"
    # TODO(desmond): Daft INTERVAL can't currently be casted correctly into timedelta objects.
    # elif dtype.is_interval():
    #     return "interval"
    elif dtype.is_list() or dtype.is_fixed_size_list():
        # TODO(desmond): We need to rework array types a little.
        # PostgreSQL supports multi-dimensional arrays and arrays of fixed sizes. However they're all
        # treated equally as a flat array and are simply documentation.
        try:
            inner_dtype = dtype.dtype
            inner_type = _daft_dtype_to_postgres_type(inner_dtype)
            # If inner type contains a complex type, simply use JSONB.
            if "[]" in inner_type or inner_type == "jsonb":
                return "jsonb"
            else:
                return f"{inner_type}[]"
        except AttributeError:
            pass
        return "jsonb"  # Fallback for unknown inner types
    elif dtype.is_struct():
        # TODO(desmond): PostgreSQL doesn't support ROW syntax in column definitions, we'd need to define new composite types.
        # Use JSONB for now.
        return "jsonb"
    elif dtype.is_map():
        return "jsonb"
    elif dtype.is_extension():
        return "jsonb"
    elif dtype.is_image() or dtype.is_fixed_shape_image():
        # TODO(desmond): Under the hood Daft uses structs for images. We should update this as we update the struct story.
        return "jsonb"
    elif dtype.is_embedding():
        # Use pgvector VECTOR type for embeddings.
        if set_dimensions:
            return f"vector({dtype.size})"
        else:
            return "vector"
    elif (
        dtype.is_tensor()
        or dtype.is_fixed_shape_tensor()
        or dtype.is_sparse_tensor()
        or dtype.is_fixed_shape_sparse_tensor()
    ):
        # TODO(desmond): Tensors are multidimensional arrays. We should update this as we update the story for lists.
        return "jsonb"
    elif dtype.is_python():
        return "jsonb"
    elif dtype.is_decimal128():
        # PostgreSQL supports decimal/numeric without parameters, or with precision,scale
        # But psycopg type registry may not know about parameterized versions
        return "numeric"
    elif dtype.is_null():
        return "text"  # Default to TEXT for null type
    else:
        return "jsonb"


def validate_connection_string(conn_string: str) -> None:
    try:
        # This will raise psycopg.ProgrammingError if invalid. See: https://www.psycopg.org/psycopg3/docs/api/conninfo.html#conninfo-manipulate-connection-strings
        psycopg.conninfo.make_conninfo(conn_string)
        return
    except psycopg.ProgrammingError as e:
        raise ValueError(f"Invalid connection string: {e}")
    except Exception as e:
        raise ValueError(f"Unexpected error: {e}")


class PostgresCatalog(Catalog):
    # TODO(desmond): For now we can create connections as needed, but in the future we can create lazy connections
    # and connection pools as an optimization.
    _inner: str  # connection string
    _extensions: list[str] | None  # extensions to create
    _create_table_options = {
        "enable_rls",
    }

    def __init__(self) -> None:
        raise RuntimeError("PostgresCatalog.__init__ is not supported, please use `Catalog.from_postgres` instead.")

    @staticmethod
    def from_uri(uri: str, extensions: list[str] | None, **options: str | None) -> PostgresCatalog:
        """Create a PostgresCatalog from a connection string."""
        validate_connection_string(uri)
        c = PostgresCatalog.__new__(PostgresCatalog)
        c._inner = uri
        c._extensions = extensions
        # TODO(desmond): Handle options.s
        return c

    @property
    def name(self) -> str:
        try:
            return psycopg.conninfo.conninfo_to_dict(self._inner)["dbname"]
        except Exception as e:
            raise ValueError(f"Error getting database name from connection string: {e}")

    ###
    # create_*
    ###

    def _create_namespace(self, identifier: Identifier) -> None:
        """Create a schema in PostgreSQL."""
        if len(identifier) != 1:
            raise ValueError(f"PostgreSQL schema identifier must be a single schema name, got {identifier}")

        quoted_schema = psycopg.sql.Identifier(identifier[0])

        with postgres_connection(self._inner, self._extensions) as conn:
            with conn.cursor() as cur:
                try:
                    cur.execute(psycopg.sql.SQL("CREATE SCHEMA {}").format(quoted_schema))
                    conn.commit()
                except psycopg.errors.DuplicateSchema:
                    raise ValueError(f"Schema {identifier} already exists")
                except psycopg.Error as e:
                    raise ValueError(f"Failed to create schema {identifier}: {e}") from e

    def _create_table(
        self,
        identifier: Identifier,
        schema: Schema,
        properties: Properties | None = None,
        partition_fields: list[PartitionField] | None = None,
    ) -> Table:
        """Create a table in PostgreSQL.

        Args:
            identifier (Identifier): The identifier of the table to create.
            schema (Schema): The schema of the table to create.
            properties (Properties): The properties of the table to create. One supported property is "enable_rls" (bool), which enables Row Level Security. This property is set to True by default. See: https://www.postgresql.org/docs/current/ddl-rowsecurity.html
            partition_fields (list[PartitionField]): The partition fields of the table to create.

        Returns:
            The table that was created.
        """
        if properties:
            self._validate_options("Postgres create_table", properties, PostgresCatalog._create_table_options)
        if len(identifier) == 1:
            # When no schema is specified, PostgreSQL uses the schema search path to select the schema to use.
            # Since this is user-configurable, we simply pass along the single identifier to PostgreSQL.
            # See: https://www.postgresql.org/docs/current/ddl-schemas.html#DDL-SCHEMAS-PATH
            schema_name = None
            table_name = identifier[0]
        elif len(identifier) == 2:
            schema_name = identifier[0]
            table_name = identifier[1]
        else:
            raise ValueError(f"PostgreSQL table identifier must be 'schema.table' or 'table', got {identifier}")

        # Build column definitions from Daft schema.
        column_defs = []
        for field in schema:
            field_name = field.name
            field_type = _daft_dtype_to_postgres_type(field.dtype)
            quoted_name = psycopg.sql.Identifier(field_name)
            column_defs.append(psycopg.sql.SQL("{} {}").format(quoted_name, psycopg.sql.SQL(field_type)))

        quoted_schema = psycopg.sql.Identifier(schema_name) if schema_name else None
        quoted_table = psycopg.sql.Identifier(table_name)
        quoted_full_table = psycopg.sql.SQL(".").join([quoted_schema, quoted_table]) if schema_name else quoted_table

        with postgres_connection(self._inner, self._extensions) as conn:
            with conn.cursor() as cur:
                try:
                    if quoted_schema:
                        cur.execute(psycopg.sql.SQL("CREATE SCHEMA IF NOT EXISTS {}").format(quoted_schema))

                    cur.execute(
                        psycopg.sql.SQL("CREATE TABLE {} ({})").format(
                            quoted_full_table, psycopg.sql.SQL(", ").join(column_defs)
                        )
                    )

                    if properties is None or properties.get("enable_rls", True):
                        cur.execute(
                            psycopg.sql.SQL("ALTER TABLE {} ENABLE ROW LEVEL SECURITY").format(quoted_full_table)
                        )

                    conn.commit()
                except psycopg.errors.DuplicateTable:
                    raise ValueError(f"Table {identifier} already exists")
                except psycopg.Error as e:
                    raise ValueError(f"Failed to create table {identifier}: {e}") from e

        return PostgresTable._from_catalog(self._inner, identifier, self._extensions)

    ###
    # drop_*
    ###

    def _drop_namespace(self, identifier: Identifier) -> None:
        """Drop a schema from PostgreSQL."""
        if len(identifier) != 1:
            raise ValueError(f"PostgreSQL namespace identifier must be a single schema name, got {identifier}")

        quoted_schema = psycopg.sql.Identifier(identifier[0])

        with postgres_connection(self._inner, self._extensions) as conn:
            with conn.cursor() as cur:
                try:
                    cur.execute(psycopg.sql.SQL("DROP SCHEMA {}").format(quoted_schema))
                    conn.commit()
                except psycopg.Error as e:
                    raise ValueError(f"Failed to drop namespace {identifier}: {e}") from e

    def _drop_table(self, identifier: Identifier) -> None:
        """Drop a table from PostgreSQL."""
        if len(identifier) == 1:
            # When no schema is specified, PostgreSQL uses the schema search path to select the schema to use.
            # Since this is user-configurable, we simply pass along the single identifier to PostgreSQL.
            # See: https://www.postgresql.org/docs/current/ddl-schemas.html#DDL-SCHEMAS-PATH
            table_name = identifier[0]
            quoted_full_table = psycopg.sql.Identifier(table_name)
        elif len(identifier) == 2:
            schema_name = identifier[0]
            table_name = identifier[1]

            quoted_schema = psycopg.sql.Identifier(schema_name)
            quoted_table = psycopg.sql.Identifier(table_name)
            quoted_full_table = psycopg.sql.SQL(".").join([quoted_schema, quoted_table])
        else:
            raise ValueError(f"PostgreSQL table identifier must be 'schema.table' or 'table', got {identifier}")

        with postgres_connection(self._inner, self._extensions) as conn:
            with conn.cursor() as cur:
                try:
                    cur.execute(psycopg.sql.SQL("DROP TABLE {}").format(quoted_full_table))
                    conn.commit()
                except psycopg.errors.UndefinedTable:
                    raise NotFoundError(f"Table {identifier} not found")
                except psycopg.Error as e:
                    raise ValueError(f"Failed to drop table {identifier}: {e}") from e

    ###
    # get_*
    ###

    def _get_table(self, identifier: Identifier) -> Table:
        """Get a table from PostgreSQL."""
        if not self._has_table(identifier):
            raise NotFoundError(f"Table {identifier} not found")

        return PostgresTable._from_catalog(self._inner, identifier, self._extensions)

    ###
    # has_*
    ###

    def _has_namespace(self, identifier: Identifier) -> bool:
        """Check if a schema exists in PostgreSQL."""
        if len(identifier) != 1:
            raise ValueError(f"PostgreSQL schema identifier must be a single schema name, got {identifier}")

        quoted_schema = psycopg.sql.Literal(identifier[0])

        with postgres_connection(self._inner, self._extensions) as conn:
            with conn.cursor() as cur:
                cur.execute(
                    psycopg.sql.SQL(
                        "select exists(SELECT 1 FROM information_schema.schemata WHERE schema_name = {})"
                    ).format(quoted_schema)
                )
                result = cur.fetchone()
                return result[0] if result else False

    def _has_table(self, identifier: Identifier) -> bool:
        """Check if a table exists in PostgreSQL."""
        if len(identifier) == 1:
            # When no schema is specified, PostgreSQL uses the schema search path to select the schema to use.
            # Since this is user-configurable, we simply pass along the single identifier to PostgreSQL.
            # See: https://www.postgresql.org/docs/current/ddl-schemas.html#DDL-SCHEMAS-PATH
            schema_name = None
            table_name = identifier[0]
        elif len(identifier) == 2:
            schema_name = identifier[0]
            table_name = identifier[1]
        else:
            return False

        with postgres_connection(self._inner, self._extensions) as conn:
            with conn.cursor() as cur:
                if schema_name:
                    cur.execute(
                        psycopg.sql.SQL(
                            "select exists(SELECT 1 FROM information_schema.tables WHERE table_schema = {} AND table_name = {})"
                        ).format(psycopg.sql.Literal(schema_name), psycopg.sql.Literal(table_name))
                    )
                else:
                    cur.execute(
                        psycopg.sql.SQL(
                            "select exists(SELECT 1 FROM information_schema.tables WHERE table_name = {})"
                        ).format(psycopg.sql.Literal(table_name))
                    )
                result = cur.fetchone()
                return result[0] if result else False

    ###
    # list_*
    ###

    def _list_namespaces(self, pattern: str | None = None) -> list[Identifier]:
        """List schemas in PostgreSQL."""
        with postgres_connection(self._inner, self._extensions) as conn:
            with conn.cursor() as cur:
                if pattern:
                    cur.execute(
                        psycopg.sql.SQL(
                            "SELECT schema_name FROM information_schema.schemata WHERE schema_name LIKE {} ORDER BY schema_name"
                        ).format(psycopg.sql.Literal(pattern + "%"))
                    )
                else:
                    cur.execute(psycopg.sql.SQL("SELECT schema_name FROM information_schema.schemata"))
                results = cur.fetchall()
                return [Identifier(row[0]) for row in results]

    def _list_tables(self, pattern: str | None = None) -> list[Identifier]:
        """List tables in PostgreSQL."""
        with postgres_connection(self._inner, self._extensions) as conn:
            with conn.cursor() as cur:
                if pattern:
                    cur.execute(
                        psycopg.sql.SQL(
                            "SELECT table_schema, table_name FROM information_schema.tables WHERE table_name LIKE {} ORDER BY table_schema, table_name"
                        ).format(psycopg.sql.Literal(pattern + "%"))
                    )
                else:
                    cur.execute(psycopg.sql.SQL("SELECT table_schema, table_name FROM information_schema.tables"))
                results = cur.fetchall()
                return [Identifier(row[0], row[1]) for row in results]


class PostgresTable(Table):
    _inner: tuple[str, Identifier]  # (connection_string, identifier)
    _extensions: list[str] | None
    _read_options = {
        "partition_col",
        "num_partitions",
        "partition_bound_strategy",
        "disable_pushdowns_to_sql",
        "infer_schema",
        "infer_schema_length",
        "schema",
    }

    def __init__(self) -> None:
        raise RuntimeError("PostgresTable.__init__ is not supported, please use `Table.from_postgres` instead.")

    @classmethod
    def _from_catalog(
        cls, connection_string: str, identifier: Identifier, extensions: list[str] | None
    ) -> PostgresTable:
        """Create a PostgresTable from catalog connection details."""
        t = cls.__new__(cls)
        t._inner = (connection_string, identifier)
        t._extensions = extensions
        return t

    @property
    def name(self) -> str:
        """Returns the table's name."""
        _, identifier = self._inner
        if len(identifier) == 1:
            # When no schema is specified, PostgreSQL uses the schema search path to select the schema to use.
            # Since this is user-configurable, we simply pass along the single identifier to PostgreSQL.
            # See: https://www.postgresql.org/docs/current/ddl-schemas.html#DDL-SCHEMAS-PATH
            return identifier[0]
        elif len(identifier) == 2:
            return identifier[1]
        else:
            raise ValueError(f"Invalid table identifier: {identifier}")

    def schema(self) -> Schema:
        """Returns the table's schema."""
        connection_string, identifier = self._inner

        if len(identifier) == 1:
            # When no schema is specified, PostgreSQL uses the schema search path to select the schema to use.
            # Since this is user-configurable, we simply pass along the single identifier to PostgreSQL.
            # See: https://www.postgresql.org/docs/current/ddl-schemas.html#DDL-SCHEMAS-PATH
            schema_name = None
            table_name = identifier[0]
        elif len(identifier) == 2:
            schema_name = identifier[0]
            table_name = identifier[1]
        else:
            raise ValueError(f"Invalid table identifier: {identifier}")

        # Query the database schema to get column information
        with postgres_connection(connection_string, self._extensions) as conn:
            with conn.cursor() as cur:
                if schema_name:
                    cur.execute(
                        psycopg.sql.SQL("""
                        SELECT
                            c.column_name,
                            c.data_type,
                            c.udt_name,
                            CASE
                                WHEN c.data_type = 'USER-DEFINED' AND c.udt_name = 'vector'
                                THEN a.atttypmod
                                ELSE NULL
                            END as vector_dimension
                        FROM information_schema.columns c
                        JOIN pg_class cls ON cls.relname = c.table_name
                        JOIN pg_namespace nsp ON nsp.oid = cls.relnamespace AND nsp.nspname = c.table_schema
                        LEFT JOIN pg_attribute a ON a.attrelid = cls.oid AND a.attname = c.column_name
                        WHERE c.table_schema = {} AND c.table_name = {}
                        ORDER BY c.ordinal_position
                        """).format(psycopg.sql.Literal(schema_name), psycopg.sql.Literal(table_name)),
                    )
                else:
                    cur.execute(
                        psycopg.sql.SQL("""
                        SELECT
                            c.column_name,
                            c.data_type,
                            c.udt_name,
                            CASE
                                WHEN c.data_type = 'USER-DEFINED' AND c.udt_name = 'vector'
                                THEN a.atttypmod
                                ELSE NULL
                            END as vector_dimension
                        FROM information_schema.columns c
                        JOIN pg_class cls ON cls.relname = c.table_name
                        LEFT JOIN pg_attribute a ON a.attrelid = cls.oid AND a.attname = c.column_name
                        WHERE c.table_name = {}
                        ORDER BY c.ordinal_position
                        """).format(psycopg.sql.Literal(table_name)),
                    )

                columns = cur.fetchall()

        # If no columns found, fall back to data-based inference
        if not columns:
            return self.read().schema()

        # Build schema from database metadata
        fields = []
        for column_name, data_type, udt_name, vector_dimension in columns:
            if data_type == "USER-DEFINED" and udt_name == "vector":
                # This is a pgvector column, convert to embedding type
                # vector_dimension from atttypmod contains the dimension information
                # For pgvector, atttypmod stores the dimension directly
                dimension = vector_dimension if vector_dimension and vector_dimension > 0 else 0

                if dimension > 0:
                    fields.append(Field.create(column_name, DataType.embedding(DataType.float32(), dimension)))
                else:
                    # Fallback to list if we can't determine dimension
                    fields.append(Field.create(column_name, DataType.list(DataType.float32())))
            else:
                # For non-vector columns, try direct PostgreSQL type mapping first
                try:
                    # Attempt to map PostgreSQL type directly to Daft type
                    inferred_dtype = DataType.from_sql(data_type)
                    fields.append(Field.create(column_name, inferred_dtype))
                except Exception:
                    # Fall back to data-based inference for unmappable types
                    # This is inefficient but ensures we get the correct types
                    if schema_name:
                        single_col_query = psycopg.sql.SQL("SELECT {} FROM {}.{} LIMIT 1").format(
                            psycopg.sql.Identifier(column_name),
                            psycopg.sql.Identifier(schema_name),
                            psycopg.sql.Identifier(table_name),
                        )
                    else:
                        single_col_query = psycopg.sql.SQL("SELECT {} FROM {} LIMIT 1").format(
                            psycopg.sql.Identifier(column_name),
                            psycopg.sql.Identifier(table_name),
                        )

                    try:
                        single_col_df = read_sql(single_col_query.as_string(), connection_string)
                        inferred_dtype = single_col_df.schema()[column_name].dtype
                        fields.append(Field.create(column_name, inferred_dtype))
                    except Exception:
                        # If inference fails, use string as fallback
                        fields.append(Field.create(column_name, DataType.string()))

        return Schema._from_fields(fields)

    @staticmethod
    def _from_obj(obj: object) -> PostgresTable:
        """Returns a PostgresTable if the given object can be adapted so."""
        raise ValueError(f"Unsupported postgres table type: {type(obj)}")

    def read(
        self,
        **options: Any,
    ) -> DataFrame:
        """Read the table as a DataFrame."""
        self._validate_options("Postgres read", options, PostgresTable._read_options)

        connection_string, identifier = self._inner

        if len(identifier) == 1:
            # When no schema is specified, PostgreSQL uses the schema search path to select the schema to use.
            # Since this is user-configurable, we simply pass along the single identifier to PostgreSQL.
            # See: https://www.postgresql.org/docs/current/ddl-schemas.html#DDL-SCHEMAS-PATH
            quoted_full_table = psycopg.sql.Identifier(identifier[0])
        elif len(identifier) == 2:
            quoted_schema = psycopg.sql.Identifier(identifier[0])
            quoted_table = psycopg.sql.Identifier(identifier[1])
            quoted_full_table = psycopg.sql.SQL(".").join([quoted_schema, quoted_table])
        else:
            raise ValueError(f"Invalid table identifier: {identifier}")

        query = psycopg.sql.SQL("SELECT * FROM {}").format(quoted_full_table)

        df = read_sql(
            query.as_string(),
            connection_string,
            **options,
        )

        # Cast any vector columns that were read as lists to embeddings
        schema = self.schema()  # Use our custom schema method
        for field in schema:
            if field.dtype.is_embedding():
                df = df.with_column(field.name, df[field.name].cast(field.dtype))

        return df

    def append(self, df: DataFrame, **options: Any) -> None:
        """Append the DataFrame to the table."""
        connection_string, identifier = self._inner

        if len(identifier) == 1:
            # When no schema is specified, PostgreSQL uses the schema search path to select the schema to use.
            # Since this is user-configurable, we simply pass along the single identifier to PostgreSQL.
            # See: https://www.postgresql.org/docs/current/ddl-schemas.html#DDL-SCHEMAS-PATH
            full_table_name = psycopg.sql.Identifier(identifier[0])
        elif len(identifier) == 2:
            quoted_schema = psycopg.sql.Identifier(identifier[0])
            quoted_table = psycopg.sql.Identifier(identifier[1])
            full_table_name = psycopg.sql.SQL(".").join([quoted_schema, quoted_table])
        else:
            raise ValueError(f"Invalid table identifier: {identifier}")

        # Build column list for COPY statement.
        columns_str = psycopg.sql.SQL(", ").join(psycopg.sql.Identifier(field.name) for field in df.schema())

        # Build types to pass into the COPY statement.
        types_to_set = [_daft_dtype_to_postgres_type(field.dtype, set_dimensions=False) for field in df.schema()]

        copy_sql = psycopg.sql.SQL("COPY {} ({}) FROM STDIN WITH (FORMAT BINARY)").format(full_table_name, columns_str)

        # TODO(desmond): This writes results sequentially on a single node. We should replace
        # this with DataFrame.write_sql() once we merge pull request #5471.
        # Additionally, although ADBC currently has limited type support, it might be worth
        # exploring for bulk loading.
        with postgres_connection(connection_string, self._extensions) as conn:
            with conn.cursor() as cur:
                with cur.copy(copy_sql) as copy:
                    copy.set_types(types_to_set)

                    for batch in df.to_arrow_iter():
                        for row in batch.to_pylist():
                            values = list(row.values())
                            # Serialize complex types that are mapped to 'text' to JSON strings for binary COPY
                            serialized_values = []
                            for value, postgres_type in zip(values, types_to_set):
                                if postgres_type in ("text", "jsonb") and not isinstance(
                                    value, (str, int, float, bool, type(None))
                                ):
                                    # Complex types (lists, dicts, etc.) need to be JSON serialized.
                                    serialized_values.append(json.dumps(value))
                                else:
                                    serialized_values.append(value)
                            copy.write_row(serialized_values)
            conn.commit()

    def overwrite(self, df: DataFrame, **options: Any) -> None:
        """Overwrite the table with the DataFrame."""
        connection_string, identifier = self._inner

        temp_catalog = PostgresCatalog.__new__(PostgresCatalog)
        temp_catalog._inner = connection_string
        temp_catalog._extensions = self._extensions

        table_exists = temp_catalog._has_table(identifier)

        if table_exists:
            temp_catalog._drop_table(identifier)

        # Extract enable_rls from options if provided, defaulting to True.
        enable_rls = options.get("enable_rls", True) if options else True
        properties = {"enable_rls": enable_rls}
        temp_catalog._create_table(identifier, df.schema(), properties=properties)
        # Now append the data (since table is empty, this effectively overwrites).
        self.append(df, **options)
        return



"""Miscellaneous Functions."""

from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Literal

import daft.daft as native
from daft.datatype import DataType, DataTypeLike
from daft.expressions import Expression, col
from daft.expressions.expressions import WhenExpr
from daft.series import item_to_series


def monotonically_increasing_id() -> Expression:
    """Generates a column of monotonically increasing unique ids.

    The implementation puts the partition number in the upper 28 bits, and the row number in each partition
    in the lower 36 bits. This allows for 2^28 ≈ 268 million partitions and 2^36 ≈ 68 billion rows per partition.

    Returns:
        Expression (UInt64 Expression): An expression that generates monotonically increasing IDs

    Examples:
        >>> import daft
        >>> from daft.functions import monotonically_increasing_id
        >>> daft.set_runner_ray()  # doctest: +SKIP
        >>>
        >>> df = daft.from_pydict({"a": [1, 2, 3, 4]}).into_partitions(2)
        >>> df = df.with_column("id", monotonically_increasing_id())
        >>> df.show()  # doctest: +SKIP
        ╭───────┬─────────────╮
        │ a     ┆ id          │
        │ ---   ┆ ---         │
        │ Int64 ┆ UInt64      │
        ╞═══════╪═════════════╡
        │ 1     ┆ 0           │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2     ┆ 1           │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 3     ┆ 68719476736 │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 4     ┆ 68719476737 │
        ╰───────┴─────────────╯
        <BLANKLINE>
        (Showing first 4 of 4 rows)

    """
    f = native.get_function_from_registry("monotonically_increasing_id")
    return Expression._from_pyexpr(f())


def uuid() -> Expression:
    """Generates a column of UUID strings.

    Each call to `uuid()` generates a fresh UUID per row. Multiple calls in the same query
    (e.g. two separate columns) are independent and will produce different values.

    Returns:
        Expression (String Expression): An expression that generates UUID strings.

    Examples:
        >>> import daft
        >>> from daft.functions import uuid
        >>> df = daft.from_pydict({"foo": [1, 2, 3]})
        >>> df = df.with_column("u1", uuid()).with_column("u2", uuid())
        >>> df.schema()["u1"].dtype == daft.DataType.string()
        True
        >>> df.schema()["u2"].dtype == daft.DataType.string()
        True
    """
    return Expression._call_builtin_scalar_fn("uuid")


def eq_null_safe(left: Expression, right: Expression) -> Expression:
    """Performs a null-safe equality comparison between two expressions.

    Unlike regular equality (==), null-safe equality (<=> or IS NOT DISTINCT FROM):
    - Returns True when comparing NULL <=> NULL
    - Returns False when comparing NULL <=> any_value
    - Behaves like regular equality for non-NULL values

    Returns:
        Expression (Boolean Expression): A boolean expression indicating if the values are equal
    """
    left = Expression._to_expression(left)
    right = Expression._to_expression(right)
    return Expression._from_pyexpr(left._expr.eq_null_safe(right._expr))


def cast(expr: Expression, dtype: DataTypeLike) -> Expression:
    """Casts an expression to the given datatype if possible.

    See the [casting matrix](https://docs.daft.ai/en/stable/api/datatypes/casting/) for supported casts.

    Returns:
        Expression: Expression with the specified new datatype

    Note:
        - If a string is provided, it will use the sql engine to parse the string into a data type. See the [SQL Reference](https://docs.daft.ai/en/stable/sql/datatypes/) for supported datatypes.
        - a python `type` can also be provided, in which case the corresponding Daft data type will be used.

    Tip: See Also
        [`Expression.cast`](https://docs.daft.ai/en/stable/api/expressions/#daft.expressions.Expression.cast)

    Examples:
        >>> import daft
        >>> df = daft.from_pydict({"float": [1.0, 2.5, None]})
        >>> df = df.select(df["float"].cast(daft.DataType.int64()))
        >>> df.show()
        ╭───────╮
        │ float │
        │ ---   │
        │ Int64 │
        ╞═══════╡
        │ 1     │
        ├╌╌╌╌╌╌╌┤
        │ 2     │
        ├╌╌╌╌╌╌╌┤
        │ None  │
        ╰───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

        Example with python type and sql types:
        >>> df = daft.from_pydict({"a": [1, 2, 3]})
        >>> df = df.select(
        ...     df["a"].cast(str).alias("str"),
        ...     df["a"].cast(int).alias("int"),
        ...     df["a"].cast(float).alias("float"),
        ...     df["a"].cast("string").alias("sql_string"),
        ...     df["a"].cast("int").alias("sql_int"),
        ...     df["a"].cast("tinyint").alias("sql_tinyint"),
        ... )
        >>> df.show()
        ╭────────┬───────┬─────────┬────────────┬─────────┬─────────────╮
        │ str    ┆ int   ┆ float   ┆ sql_string ┆ sql_int ┆ sql_tinyint │
        │ ---    ┆ ---   ┆ ---     ┆ ---        ┆ ---     ┆ ---         │
        │ String ┆ Int64 ┆ Float64 ┆ String     ┆ Int32   ┆ Int8        │
        ╞════════╪═══════╪═════════╪════════════╪═════════╪═════════════╡
        │ 1      ┆ 1     ┆ 1       ┆ 1          ┆ 1       ┆ 1           │
        ├╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2      ┆ 2     ┆ 2       ┆ 2          ┆ 2       ┆ 2           │
        ├╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 3      ┆ 3     ┆ 3       ┆ 3          ┆ 3       ┆ 3           │
        ╰────────┴───────┴─────────┴────────────┴─────────┴─────────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    dtype = DataType._infer(dtype)
    expr = Expression._to_expression(expr)
    return Expression._from_pyexpr(expr._expr.cast(dtype._dtype))


def is_null(expr: Expression) -> Expression:
    """Checks if values in the Expression are Null (a special value indicating missing data).

    Returns:
        Expression (Boolean Expression): expression indicating whether values are missing

    Examples:
        >>> import daft
        >>> from daft.functions import is_null
        >>>
        >>> df = daft.from_pydict({"x": [1.0, None, float("nan")]})
        >>> df = df.select(is_null(df["x"]))
        >>> df.collect()
        ╭───────╮
        │ x     │
        │ ---   │
        │ Bool  │
        ╞═══════╡
        │ false │
        ├╌╌╌╌╌╌╌┤
        │ true  │
        ├╌╌╌╌╌╌╌┤
        │ false │
        ╰───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    expr = Expression._to_expression(expr)
    return Expression._from_pyexpr(expr._expr.is_null())


def not_null(expr: Expression) -> Expression:
    """Checks if values in the Expression are not Null (a special value indicating missing data).

    Returns:
        Expression (Boolean Expression): expression indicating whether values are not missing

    Examples:
        >>> import daft
        >>> from daft.functions import not_null
        >>>
        >>> df = daft.from_pydict({"x": [1.0, None, float("nan")]})
        >>> df = df.select(not_null(df["x"]))
        >>> df.collect()
        ╭───────╮
        │ x     │
        │ ---   │
        │ Bool  │
        ╞═══════╡
        │ true  │
        ├╌╌╌╌╌╌╌┤
        │ false │
        ├╌╌╌╌╌╌╌┤
        │ true  │
        ╰───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    expr = Expression._to_expression(expr)
    return Expression._from_pyexpr(expr._expr.not_null())


def fill_null(expr: Expression, fill_value: Expression) -> Expression:
    """Fills null values in the Expression with the provided fill_value.

    Returns:
        Expression: Expression with null values filled with the provided fill_value

    Examples:
        >>> import daft
        >>> from daft.functions import fill_null
        >>>
        >>> df = daft.from_pydict({"data": [1, None, 3]})
        >>> df = df.select(fill_null(df["data"], 2))
        >>> df.collect()
        ╭───────╮
        │ data  │
        │ ---   │
        │ Int64 │
        ╞═══════╡
        │ 1     │
        ├╌╌╌╌╌╌╌┤
        │ 2     │
        ├╌╌╌╌╌╌╌┤
        │ 3     │
        ╰───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    expr = Expression._to_expression(expr)
    fill_value = Expression._to_expression(fill_value)
    return Expression._from_pyexpr(expr._expr.fill_null(fill_value._expr))


def is_in(expr: Expression, other: Iterable[Any] | Expression) -> Expression:
    """Checks if values in the Expression are in the provided iterable.

    Args:
        expr: The expression to check
        other: An iterable (list, set, tuple, etc.), Expression, or array-like object containing the values to check against

    Returns:
        Expression (Boolean Expression): expression indicating whether values are in the provided iterable

    Examples:
        >>> import daft
        >>> from daft.functions import is_in
        >>>
        >>> df = daft.from_pydict({"data": [1, 2, 3]})
        >>> df = df.select(is_in(df["data"], [1, 3]))
        >>> df.collect()
        ╭───────╮
        │ data  │
        │ ---   │
        │ Bool  │
        ╞═══════╡
        │ true  │
        ├╌╌╌╌╌╌╌┤
        │ false │
        ├╌╌╌╌╌╌╌┤
        │ true  │
        ╰───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    # Convert non-list iterables (sets, tuples, generators, ranges, etc.) to lists
    # Exclude strings/bytes since they are iterable but should not be treated as sequences of characters/bytes
    if isinstance(other, Iterable) and not isinstance(other, (str, bytes, Expression)):
        other = list(other)

    if isinstance(other, list):
        other = [Expression._to_expression(item) for item in other]
    elif not isinstance(other, Expression):
        series = item_to_series("items", other)
        other = [Expression._from_pyexpr(native.list_lit(series._series))]
    else:
        other = [other]

    expr = Expression._to_expression(expr)
    return Expression._from_pyexpr(expr._expr.is_in([item._expr for item in other]))


def hash(
    *exprs: Expression,
    seed: Any | None = None,
    hash_function: Literal["xxhash", "xxhash32", "xxhash64", "xxhash3_64", "murmurhash3", "sha1"] | None = "xxhash",
) -> Expression:
    """Hashes the values in the Expression.

    Uses the specified hash function to hash the values in the expression. Default to [XXH3_64bits](https://xxhash.com/) non-cryptographic hash function.

    Args:
        exprs: One or more expressions (or column names/wildcards) to hash.
        seed (optional): Seed used for generating the hash. Defaults to 0.
        hash_function (optional): Hash function to use. One of "xxhash" (alias for "xxhash3_64"), "xxhash32", "xxhash64", "xxhash3_64", "murmurhash3", or "sha1". Defaults to "xxhash" (alias for "xxhash3_64").

    Returns:
        Expression (UInt64 Expression): The hashed expression.

    Note:
        Null values will produce a hash value instead of being propagated as null.

    """
    # Only pass hash_function if explicitly provided to maintain backward compatibility in string representation
    kwargs = {}
    if not exprs:
        raise ValueError("hash() requires at least one expression")
    normalized = []
    for expr in exprs:
        if isinstance(expr, str):
            resolved = col(expr)
            normalized.append(resolved)
        else:
            resolved = Expression._to_expression(expr)
            normalized.append(resolved)
    if seed is not None:
        kwargs["seed"] = seed
    if hash_function is not None:
        kwargs["hash_function"] = hash_function
    return Expression._call_builtin_scalar_fn("hash", *normalized, **kwargs)


def minhash(
    text: Expression,
    *,
    num_hashes: int,
    ngram_size: int,
    seed: int = 1,
    hash_function: Literal["murmurhash3", "xxhash", "xxhash32", "xxhash64", "xxhash3_64", "sha1"] = "murmurhash3",
) -> Expression:
    """Runs the MinHash algorithm on the series.

    For a string, calculates the minimum hash over all its ngrams,
    repeating with `num_hashes` permutations. Returns as a list of 32-bit unsigned integers.

    Tokens for the ngrams are delimited by spaces.
    The strings are not normalized or pre-processed, so it is recommended
    to normalize the strings yourself.

    Args:
        text (String Expression): The expression to hash.
        num_hashes (int): The number of hash permutations to compute.
        ngram_size (int): The number of tokens in each shingle/ngram.
        seed (int, default=1): Seed used for generating permutations and the initial string hashes. Defaults to 1.
        hash_function (str, default="murmurhash3"): Hash function to use for initial string hashing. One of "murmurhash3", "xxhash" (alias for "xxhash3_64"), "xxhash32", "xxhash64", "xxhash3_64", or "sha1". Defaults to "murmurhash3".

    Returns:
        Expression (FixedSizedList[UInt32, num_hashes] Expression):
            expression representing the MinHash values.

    """
    return Expression._call_builtin_scalar_fn(
        "minhash", text, num_hashes=num_hashes, ngram_size=ngram_size, seed=seed, hash_function=hash_function
    )


def length(expr: Expression) -> Expression:
    """Retrieves the length of the given expression.

    Args:
        expr (List or Binary or String Expression): expression to compute the length of.

    The behavior depends on the input type:
    - For strings, returns the number of characters.
    - For binary, returns the number of bytes.
    - For lists, returns the number of elements.

    Returns:
        Expression (UInt64 Expression): an expression with the length

    Examples:
        String length:
        >>> import daft
        >>> from daft.functions import length
        >>>
        >>> df = daft.from_pydict({"x": ["foo", "bar", None]})
        >>> df = df.select(length(df["x"]))
        >>> df.show()
        ╭────────╮
        │ x      │
        │ ---    │
        │ UInt64 │
        ╞════════╡
        │ 3      │
        ├╌╌╌╌╌╌╌╌┤
        │ 3      │
        ├╌╌╌╌╌╌╌╌┤
        │ None   │
        ╰────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

        Binary length:
        >>> df = daft.from_pydict({"x": [b"foo", b"bar", None]})
        >>> df = df.select(length(df["x"]))
        >>> df.show()
        ╭────────╮
        │ x      │
        │ ---    │
        │ UInt64 │
        ╞════════╡
        │ 3      │
        ├╌╌╌╌╌╌╌╌┤
        │ 3      │
        ├╌╌╌╌╌╌╌╌┤
        │ None   │
        ╰────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

        List length:
        >>> df = daft.from_pydict({"x": [[1, 2, 3], [4, 5], None]})
        >>> df = df.select(length(df["x"]))
        >>> df.show()
        ╭────────╮
        │ x      │
        │ ---    │
        │ UInt64 │
        ╞════════╡
        │ 3      │
        ├╌╌╌╌╌╌╌╌┤
        │ 2      │
        ├╌╌╌╌╌╌╌╌┤
        │ None   │
        ╰────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    return Expression._call_builtin_scalar_fn("length", expr)


def concat(left: Expression | str | bytes, right: Expression | str | bytes) -> Expression:
    r"""Concatenates two string or binary values.

    Args:
        left ((String or Binary Expression) | str | bytes): the left value to concatenate
        right ((String or Binary Expression) | str | bytes): the right value to concatenate

    Returns:
        Expression: an expression with the same type as the inputs

    Examples:
        String concatenation:

        >>> import daft
        >>> from daft.functions import concat
        >>>
        >>> df = daft.from_pydict({"x": ["foo", "bar", "baz"], "y": ["a", "b", "c"]})
        >>> df.select(concat(df["x"], df["y"])).collect()
        ╭────────╮
        │ x      │
        │ ---    │
        │ String │
        ╞════════╡
        │ fooa   │
        ├╌╌╌╌╌╌╌╌┤
        │ barb   │
        ├╌╌╌╌╌╌╌╌┤
        │ bazc   │
        ╰────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

        Binary concatenation:

        >>> df = daft.from_pydict(
        ...     {"a": [b"Hello", b"\\xff\\xfe", b"", b"World"], "b": [b" World", b"\\x00", b"empty", b"!"]}
        ... )
        >>> df = df.select(concat(df["a"], df["b"]))
        >>> df.show()
        ╭────────────────────╮
        │ a                  │
        │ ---                │
        │ Binary             │
        ╞════════════════════╡
        │ b"Hello World"     │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ b"\\xff\\xfe\\x00" │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ b"empty"           │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ b"World!"          │
        ╰────────────────────╯
        <BLANKLINE>
        (Showing first 4 of 4 rows)

    """
    return Expression._to_expression(left) + Expression._to_expression(right)


def coalesce(*args: Expression) -> Expression:
    """Returns the first non-null value in a list of expressions. If all inputs are null, returns null.

    Args:
        *args: Two or more expressions to coalesce

    Returns:
        Expression: Expression containing first non-null value encountered when evaluating arguments in order

    Examples:
        >>> import daft
        >>> from daft.functions import coalesce
        >>> df = daft.from_pydict({"x": [1, None, 3], "y": [None, 2, None]})
        >>> df = df.with_column("first_valid", coalesce(df["x"], df["y"]))
        >>> df.show()
        ╭───────┬───────┬─────────────╮
        │ x     ┆ y     ┆ first_valid │
        │ ---   ┆ ---   ┆ ---         │
        │ Int64 ┆ Int64 ┆ Int64       │
        ╞═══════╪═══════╪═════════════╡
        │ 1     ┆ None  ┆ 1           │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ None  ┆ 2     ┆ 2           │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 3     ┆ None  ┆ 3           │
        ╰───────┴───────┴─────────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("coalesce", *args)


def get(expr: Expression, key: int | str | Expression, default: Any = None) -> Expression:
    """Get an index from a list expression or a field from a struct expression.

    Args:
        expr (List or Struct Expression): to get value from
        key: integer index for list or string field for struct. List index can be negative to index from the end of the list.
        default: default value if out of bounds. Only supported for list get

    Returns:
        An expression with the inner type of the input expression.

    Note:
        `expr.get(x)` can also be written as `expr[x]`

    Note:
        `expr.get("*")` is equivalent to `expr.unnest()`

    Examples:
        Getting elements from a list by index:

        >>> import daft
        >>> df = daft.from_pydict({"lists": [[1, 2, 3], [4, 5], [6]]})
        >>> df = df.select(df["lists"].get(0).alias("first"), df["lists"].get(-1).alias("last"))
        >>> df.show()
        ╭───────┬───────╮
        │ first ┆ last  │
        │ ---   ┆ ---   │
        │ Int64 ┆ Int64 │
        ╞═══════╪═══════╡
        │ 1     ┆ 3     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 4     ┆ 5     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 6     ┆ 6     │
        ╰───────┴───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

        Getting elements from a list with default value:

        >>> df = daft.from_pydict({"lists": [[1, 2], [3], []]})
        >>> df = df.select(df["lists"].get(2, default=-1))
        >>> df.show()
        ╭───────╮
        │ lists │
        │ ---   │
        │ Int64 │
        ╞═══════╡
        │ -1    │
        ├╌╌╌╌╌╌╌┤
        │ -1    │
        ├╌╌╌╌╌╌╌┤
        │ -1    │
        ╰───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

        Getting fields from a struct:

        >>> df = daft.from_pydict({"structs": [{"name": "Alice", "age": 25}, {"name": "Bob", "age": 30}]})
        >>> df = df.select(df["structs"].get("name"), df["structs"].get("age"))
        >>> df.show()
        ╭────────┬───────╮
        │ name   ┆ age   │
        │ ---    ┆ ---   │
        │ String ┆ Int64 │
        ╞════════╪═══════╡
        │ Alice  ┆ 25    │
        ├╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ Bob    ┆ 30    │
        ╰────────┴───────╯
        <BLANKLINE>
        (Showing first 2 of 2 rows)

        Using variable indices:

        >>> df = daft.from_pydict({"lists": [[1, 2, 3], [4, 5, 6]], "indices": [0, 2]})
        >>> df = df.select(df["lists"].get(df["indices"]))
        >>> df.show()
        ╭───────╮
        │ lists │
        │ ---   │
        │ Int64 │
        ╞═══════╡
        │ 1     │
        ├╌╌╌╌╌╌╌┤
        │ 6     │
        ╰───────╯
        <BLANKLINE>
        (Showing first 2 of 2 rows)

        Unnesting all fields from a struct (equivalent to .unnest()):

        >>> df = daft.from_pydict({"structs": [{"x": 1, "y": 2}, {"x": 3, "y": 4}]})
        >>> df = df.select(df["structs"].get("*"))
        >>> df.show()
        ╭───────┬───────╮
        │ x     ┆ y     │
        │ ---   ┆ ---   │
        │ Int64 ┆ Int64 │
        ╞═══════╪═══════╡
        │ 1     ┆ 2     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 3     ┆ 4     │
        ╰───────┴───────╯
        <BLANKLINE>
        (Showing first 2 of 2 rows)
    """
    if isinstance(key, (int, Expression)):
        return Expression._call_builtin_scalar_fn("list_get", expr, key, default)
    elif isinstance(key, str):
        if default is not None:
            raise ValueError("`daft.functions.get` does not support default values for getting a struct field")
        return Expression._from_pyexpr(expr._expr.struct_get(key))
    else:
        raise TypeError(
            f"Argument {key} of type {type(key)} is not supported in `daft.functions.get`. Only int and string types are supported."
        )


def map_get(expr: Expression, key: Expression) -> Expression:
    """Retrieves the value for a key in a map column.

    Args:
        expr: the map expression to get from
        key: the key to retrieve

    Returns:
        Expression: the value expression

    Examples:
        >>> import pyarrow as pa
        >>> import daft
        >>> pa_array = pa.array([[("a", 1)], [], [("b", 2)]], type=pa.map_(pa.string(), pa.int64()))
        >>> df = daft.from_arrow(pa.table({"map_col": pa_array}))
        >>> df = df.with_column("a", df["map_col"].map_get("a"))
        >>> df.show()
        ╭────────────────────┬───────╮
        │ map_col            ┆ a     │
        │ ---                ┆ ---   │
        │ Map[String: Int64] ┆ Int64 │
        ╞════════════════════╪═══════╡
        │ {"a": 1}           ┆ 1     │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ {}                 ┆ None  │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ {"b": 2}           ┆ None  │
        ╰────────────────────┴───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    key_expr = Expression._to_expression(key)
    return Expression._from_pyexpr(expr._expr.map_get(key_expr._expr))


def slice(expr: Expression, start: int | Expression, end: int | Expression | None = None) -> Expression:
    r"""Get a subset of each list or binary value.

    Args:
        expr: List or binary expression to slice.
        start: Index or column of indices. The slice will include elements starting from this index. If `start` is negative, it represents an offset from the end
        end: Index or column of indices. The slice will not include elements from this index onwards. If `end` is negative, it represents an offset from the end. If not provided, the slice will include elements up to the end of the list. If start > end, an empty slice is produced.

    Returns:
        Expression: an expression with the same type as the input.

    Note:
        `expr[start:stop]` is also equivalent to `expr.slice(start, stop)`

    Examples:
        Slicing a list expression:
        >>> import daft
        >>> df = daft.from_pydict({"x": [[1, 2, 3], [4, 5, 6, 7], [8]]})
        >>> df = df.select(df["x"].slice(1, -1))
        >>> df.show()
        ╭─────────────╮
        │ x           │
        │ ---         │
        │ List[Int64] │
        ╞═════════════╡
        │ [2]         │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ [5, 6]      │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ []          │
        ╰─────────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

        Slicing a binary expression:
        >>> df = daft.from_pydict({"x": [b"Hello World", b"\xff\xfe\x00", b"empty"]})
        >>> df = df.select(df["x"].slice(1, -2))
        >>> df.show()
        ╭─────────────╮
        │ x           │
        │ ---         │
        │ Binary      │
        ╞═════════════╡
        │ b"ello Wor" │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ b""         │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ b"mp"       │
        ╰─────────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    return Expression._call_builtin_scalar_fn("slice", expr, start, end=end)


def when(condition: Expression | bool, then: Expression | Any) -> WhenExpr:
    """Start a conditional expression, similar to SQL CASE WHEN.

    If the condition is true, the `then` value will be returned. Otherwise, the next `when` condition will be evaluated.
    If no conditions are true, the value will be set to the value provided in the `otherwise` clause, or null if not provided.

    Args:
        condition: The Boolean expression to evaluate
        then: Expression to return when the condition is true

    Returns:
        A WhenExpr that can be chained with more `when` clauses and ended with `otherwise`

    Examples:
        Simple conditional assignment:
        >>> import daft
        >>> from daft.functions import when
        >>>
        >>> df = daft.from_pydict({"x": [1, 2, 3, 4, 5]})
        >>> df = df.select(when(df["x"] > 3, then="high").otherwise("low").alias("category"))
        >>> df.show()
        ╭──────────╮
        │ category │
        │ ---      │
        │ String   │
        ╞══════════╡
        │ low      │
        ├╌╌╌╌╌╌╌╌╌╌┤
        │ low      │
        ├╌╌╌╌╌╌╌╌╌╌┤
        │ low      │
        ├╌╌╌╌╌╌╌╌╌╌┤
        │ high     │
        ├╌╌╌╌╌╌╌╌╌╌┤
        │ high     │
        ╰──────────╯
        <BLANKLINE>
        (Showing first 5 of 5 rows)

        Multiple conditions using chained `when` clauses:
        >>> df = daft.from_pydict({"score": [85, 92, 78, 65, 88]})
        >>> df = df.select(
        ...     when(df["score"] >= 90, then="A")
        ...     .when(df["score"] >= 80, then="B")
        ...     .when(df["score"] >= 70, then="C")
        ...     .otherwise("F")
        ...     .alias("grade")
        ... )
        >>> df.show()
        ╭────────╮
        │ grade  │
        │ ---    │
        │ String │
        ╞════════╡
        │ B      │
        ├╌╌╌╌╌╌╌╌┤
        │ A      │
        ├╌╌╌╌╌╌╌╌┤
        │ C      │
        ├╌╌╌╌╌╌╌╌┤
        │ F      │
        ├╌╌╌╌╌╌╌╌┤
        │ B      │
        ╰────────╯
        <BLANKLINE>
        (Showing first 5 of 5 rows)

        Using complex conditions and returning different data types:
        >>> df = daft.from_pydict({"name": ["Alice", "Bob", "Charlie"], "age": [25, 17, 35]})
        >>> df = df.select(
        ...     df["name"],
        ...     when((df["age"] >= 18) & (df["age"] < 65), then=df["age"])
        ...     .when(df["age"] < 18, then=-1)
        ...     .otherwise(0)
        ...     .alias("working_age"),
        ... )
        >>> df.show()
        ╭─────────┬─────────────╮
        │ name    ┆ working_age │
        │ ---     ┆ ---         │
        │ String  ┆ Int64       │
        ╞═════════╪═════════════╡
        │ Alice   ┆ 25          │
        ├╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ Bob     ┆ -1          │
        ├╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ Charlie ┆ 35          │
        ╰─────────┴─────────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

        Handling null values:
        >>> df = daft.from_pydict({"value": [10, None, 20, 0]})
        >>> df = df.select(
        ...     when(df["value"].is_null(), then="missing")
        ...     .when(df["value"] == 0, then="zero")
        ...     .when(df["value"] > 15, then="high")
        ...     .otherwise("normal")
        ...     .alias("status")
        ... )
        >>> df.show()
        ╭─────────╮
        │ status  │
        │ ---     │
        │ String  │
        ╞═════════╡
        │ normal  │
        ├╌╌╌╌╌╌╌╌╌┤
        │ missing │
        ├╌╌╌╌╌╌╌╌╌┤
        │ high    │
        ├╌╌╌╌╌╌╌╌╌┤
        │ zero    │
        ╰─────────╯
        <BLANKLINE>
        (Showing first 4 of 4 rows)

        Without `otherwise` clause (returns null when no conditions match):
        >>> df = daft.from_pydict({"x": [1, 2, 3]})
        >>> df = df.select(when(df["x"] > 1, then="big").alias("result"))
        >>> df.show()
        ╭────────╮
        │ result │
        │ ---    │
        │ String │
        ╞════════╡
        │ None   │
        ├╌╌╌╌╌╌╌╌┤
        │ big    │
        ├╌╌╌╌╌╌╌╌┤
        │ big    │
        ╰────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    return WhenExpr([]).when(condition, then)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, TypedDict

from daft.ai.protocols import Prompter, PrompterDescriptor

if TYPE_CHECKING:
    from daft.ai.typing import Options


class VLLMPromptOptions(TypedDict, total=False):
    concurrency: int
    gpus_per_actor: int
    do_prefix_routing: bool
    max_buffer_size: int
    min_bucket_size: int
    prefix_match_threshold: float
    load_balance_threshold: int
    batch_size: int | None
    engine_args: dict[str, Any]
    generate_args: dict[str, Any]
    num_gpus: int


@dataclass
class VLLMPrefixCachingPrompterDescriptor(PrompterDescriptor):
    """Descriptor for vLLM prompter.

    Note: This descriptor is not actually used to instantiate a Prompter.
    Instead, the prompt() function detects this descriptor type and calls
    PyExpr.vllm() directly for optimized prefix-cached inference.
    """

    provider_name: str
    options: VLLMPromptOptions
    return_format: Any | None = None
    system_message: str | None = None
    model_name: str = "facebook/opt-125m"

    def get_provider(self) -> str:
        return self.provider_name

    def get_model(self) -> str:
        return self.model_name

    def get_options(self) -> Options:
        defaults: VLLMPromptOptions = {
            "concurrency": 1,
            "gpus_per_actor": 1,
            "do_prefix_routing": True,
            "max_buffer_size": 5000,
            "min_bucket_size": 16,
            "prefix_match_threshold": 0.33,
            "load_balance_threshold": 256,
            "batch_size": None,
            "engine_args": {},
            "generate_args": {},
            "num_gpus": 1,
        }
        return {**defaults, **self.options}

    def instantiate(self) -> Prompter:
        """This method should not be called for vLLM provider.

        The prompt() function bypasses the UDF execution path for vLLM
        and calls PyExpr.vllm() directly instead.
        """
        raise NotImplementedError(
            "VLLMPrompterDescriptor.instantiate() should not be called. "
            "The prompt() function uses PyExpr.vllm() directly for vLLM provider."
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from daft.daft import (
    Input,
    LocalPhysicalPlan,
    PyDaftExecutionConfig,
    PyExecutionStats,
    PyMicroPartition,
)
from daft.daft import (
    NativeExecutor as _NativeExecutor,
)
from daft.dataframe.display import MermaidOptions
from daft.event_loop import get_or_init_event_loop
from daft.recordbatch import MicroPartition
from daft.runners.partitioning import LocalMaterializedResult

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Generator, Mapping

    from daft.context import DaftContext
    from daft.logical.builder import LogicalPlanBuilder


class NativeExecutor:
    def __init__(self) -> None:
        self._executor = _NativeExecutor()

    def run(
        self,
        local_physical_plan: LocalPhysicalPlan,
        inputs: Mapping[int, Input | list[PyMicroPartition]],
        ctx: DaftContext,
        context: dict[str, str] | None,
    ) -> Generator[LocalMaterializedResult, None, tuple[str, PyExecutionStats]]:
        stats: PyExecutionStats | None = None
        query_plan: str | None = None

        async def stream_results() -> AsyncGenerator[PyMicroPartition | None, None]:
            result_handle = await self._executor.run(
                local_physical_plan,
                ctx._ctx,
                0,
                dict(inputs),
                context,
            )
            nonlocal query_plan
            query_plan = await result_handle.query_plan()
            nonlocal stats
            try:
                async for batch in result_handle:
                    yield batch
            finally:
                stats = await result_handle.finish()

        event_loop = get_or_init_event_loop()
        async_exec = stream_results()
        should_raise_errors_from_close = True
        try:
            while True:
                part = event_loop.run(async_exec.__anext__())
                if part is None:
                    break
                yield LocalMaterializedResult(MicroPartition._from_pymicropartition(part))
        except BaseException:
            # Preserve the original exception/GeneratorExit by not masking it with errors from async_exec.aclose().
            should_raise_errors_from_close = False
            raise
        finally:
            try:
                event_loop.run(async_exec.aclose())
            except Exception:
                if should_raise_errors_from_close:
                    raise

        assert query_plan is not None and stats is not None
        return query_plan, stats

    def pretty_print(
        self,
        builder: LogicalPlanBuilder,
        daft_execution_config: PyDaftExecutionConfig,
        simple: bool = False,
        format: str = "ascii",
    ) -> str:
        """Pretty prints the current underlying logical plan."""
        if format == "ascii":
            return _NativeExecutor.repr_ascii(builder._builder, daft_execution_config, simple)
        elif format == "mermaid":
            return _NativeExecutor.repr_mermaid(builder._builder, daft_execution_config, MermaidOptions(simple))
        else:
            raise ValueError(f"Unknown format: {format}")



"""Process Functions."""

from __future__ import annotations

import subprocess
from typing import Any, Literal

import daft
from daft.datatype import DataType, DataTypeLike
from daft.expressions import Expression


def run_process(
    args: Expression | list[Expression | Any],
    *,
    shell: bool = False,
    on_error: Literal["raise", "ignore", "log"] = "log",
    return_dtype: DataTypeLike = DataType.string(),
) -> Expression:
    """Returns an expression that runs an external process (optionally via a shell) and exposes its stdout as a column.

    This helper wraps a Python UDF around ``subprocess.run`` so it can be used inside DataFrame expressions.

    Args:
        args (Expression | list[Expression | Any]):
            The command to execute.
            If ``shell=False`` (default), pass a list of arguments, for example ``["ls", "-a", col("path")]``.
            If ``shell=True``, pass a single string expression or a list that will be joined, for example ``"echo hello"`` or ``["echo", "hello"]``.
        shell (bool, default=False):
            Whether to execute the command via the system shell (equivalent to ``subprocess.run(..., shell=True)``).
            Using the shell enables pipes and redirection but is more vulnerable to injection. Defaults to ``False``.
        on_error (Literal["raise", "ignore", "log"], default="log"):
            Whether to log an error when encountering an error, or log a warning and return a null
        return_dtype: Desired Daft data type for the result column. Defaults to a UTF-8 string column.

    Returns:
        Expression: An expression representing the stdout of the process converted to ``return_dtype`` (defaults to a UTF-8 string column).

    Examples:
        >>> import daft
        >>> from daft import col
        >>> from daft.functions import run_process
        >>> df = daft.from_pydict({"a": ["hello"], "b": ["world"]})
        >>> # Without shell
        >>> expr = run_process(["echo", col("a"), col("b")])
        >>> df = df.select(expr.alias("out"))
        >>> df.to_pylist()[0]["out"].strip()
        'hello world'
        >>>
        >>> # With shell and return_dtype=int
        >>> df = daft.from_pydict({"x": ["hello world"]})
        >>> expr = run_process(("echo " + col("x") + " | wc -c"), shell=True, return_dtype=int)
        >>> df = df.select(expr.alias("word_count"))
        >>> df.to_pylist()
        [{'word_count': 12}]
    """

    @daft.func(return_dtype=return_dtype, on_error=on_error)
    def _impl(*argv: Any) -> Any:
        if shell:
            if len(argv) != 1:
                raise ValueError("shell=True expects a single string expression as the command")
            cmd_str = str(argv[0])
            proc = subprocess.run(cmd_str, shell=True, stdout=subprocess.PIPE, text=True, check=True)
        else:
            if len(argv) == 0:
                raise ValueError("shell=False requires at least one argv token")
            tokens = [str(a) for a in argv]
            proc = subprocess.run(tokens, shell=False, stdout=subprocess.PIPE, text=True, check=True)
        return proc.stdout

    if isinstance(args, Expression):
        args = [args]
    elif not isinstance(args, list):
        args = [args]

    expr_args = [Expression._to_expression(v) for v in args]
    return _impl(*expr_args).cast(return_dtype)



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from functools import cache
from typing import TYPE_CHECKING

from huggingface_hub.hf_api import HfApi

from daft.datatype import DataType
from daft.dependencies import pq
from daft.io.sink import DataSink, WriteResult
from daft.recordbatch.micropartition import MicroPartition
from daft.schema import Schema
from daft.utils import get_arrow_version

if TYPE_CHECKING:
    from collections.abc import Iterator

    from huggingface_hub import CommitOperationAdd, HfApi

    from daft.daft import HuggingFaceConfig


@dataclass
class CommitOperationAddWrapper:
    """Wrapper class to prevent the need for `CommitOperationAdd` to be imported outside of the type checking block.

    It needs to be available outside of the type checking block because it's used as a generic parameter and thus requires the type to be available at runtime.

    More details here: https://github.com/astral-sh/ruff/issues/7214
    """

    inner: CommitOperationAdd


class HuggingFaceSink(DataSink[CommitOperationAddWrapper]):
    def __init__(
        self,
        repo: str,
        split: str,
        data_dir: str,
        revision: str,
        overwrite: bool,
        commit_message: str,
        commit_description: str | None,
        config: HuggingFaceConfig,
    ):
        self.repo = repo
        self.split = split
        self.data_dir = data_dir
        self.revision = revision
        self.overwrite = overwrite
        self.commit_message = commit_message
        self.commit_description = commit_description
        self.config = config

        from importlib.util import find_spec

        if find_spec("huggingface_hub") is None:
            raise ImportError(
                "huggingface-hub must be installed to run `write_huggingface`. Install it with `pip install -U 'daft[huggingface]'`"
            )

        if self.config.anonymous:
            raise ValueError("Cannot write to Hugging Face dataset in anonymous mode.")

        self.path_prefix = f"{self.data_dir}/{self.split}".strip("/")

        if get_arrow_version() < (21, 0, 0):
            if self.config.use_content_defined_chunking:
                import warnings

                warnings.warn(
                    "HuggingFaceConfig.use_content_defined_chunking is enabled but requires PyArrow >=21.0.0. This setting will be ignored."
                )

            self.writer_kwargs = {}
        else:
            self.writer_kwargs = {
                "use_content_defined_chunking": bool(self.config.use_content_defined_chunking),
            }

    def name(self) -> str:
        return "Hugging Face Dataset Write"

    @cache
    def schema(self) -> Schema:
        return Schema.from_pydict(
            {
                "path_in_repo": DataType.string(),
                "operation": DataType.string(),
                "src_path_in_repo": DataType.string(),
            }
        )

    @contextmanager
    def _hf_api(self) -> Iterator[HfApi]:
        """Context manager to get an authenticated Hugging Face API client with progress bars disabled."""
        from huggingface_hub import HfApi
        from huggingface_hub.utils import disable_progress_bars, enable_progress_bars

        api = HfApi(token=self.config.token, library_name="daft")
        disable_progress_bars()
        try:
            yield api
        finally:
            enable_progress_bars()

    def write(self, micropartitions: Iterator[MicroPartition]) -> Iterator[WriteResult[CommitOperationAddWrapper]]:
        import io
        import uuid

        from huggingface_hub import CommitOperationAdd

        id = uuid.uuid4()
        num_files = 0
        num_rows = 0

        with self._hf_api() as api:
            with io.BytesIO() as buffer:
                writer = None

                def flush(writer: pq.ParquetWriter) -> WriteResult[CommitOperationAddWrapper]:
                    writer.close()

                    nonlocal num_files
                    nonlocal num_rows

                    path_in_repo = f"{self.path_prefix}-{id}-{num_files}.parquet"
                    num_files += 1

                    addition = CommitOperationAdd(path_in_repo=path_in_repo, path_or_fileobj=buffer.getvalue())
                    api.preupload_lfs_files(
                        repo_id=self.repo, additions=[addition], repo_type="dataset", revision=self.revision
                    )

                    # Reuse the buffer for the next file
                    buffer.seek(0)
                    buffer.truncate()

                    result = WriteResult(
                        CommitOperationAddWrapper(addition),
                        bytes_written=addition.upload_info.size,
                        rows_written=num_rows,
                    )

                    num_rows = 0

                    return result

                for part in micropartitions:
                    if writer is None:
                        writer = pq.ParquetWriter(buffer, part.schema().to_pyarrow_schema(), **self.writer_kwargs)

                    num_rows += len(part)
                    writer.write_table(part.to_arrow(), row_group_size=self.config.row_group_size)

                    if buffer.tell() > self.config.target_filesize:
                        yield flush(writer)

                        # delete the writer and create one for the next partition
                        writer = None

                # flush remaining data
                if writer is not None:
                    yield flush(writer)

    def finalize(self, write_results: list[WriteResult[CommitOperationAddWrapper]]) -> MicroPartition:
        from huggingface_hub import CommitOperation, CommitOperationAdd, CommitOperationCopy, CommitOperationDelete
        from huggingface_hub.hf_api import RepoFile, RepoFolder

        additions = [res.result.inner for res in write_results]
        operations = {}
        count_new = len(additions)
        count_existing = 0
        output: dict[str, list[str]] = {"path_in_repo": [], "operation": [], "src_path_in_repo": []}

        def list_split(api: HfApi) -> Iterator[RepoFile | RepoFolder]:
            """Get all existing files of the current split."""
            from huggingface_hub.utils import EntryNotFoundError

            try:
                objects = api.list_repo_tree(
                    path_in_repo=self.data_dir,
                    repo_id=self.repo,
                    repo_type="dataset",
                    revision=self.revision,
                    expand=False,
                    recursive=False,
                )
                for obj in objects:
                    if obj.path.startswith(self.path_prefix):
                        yield obj
            except EntryNotFoundError:
                pass

        def create_commits(api: HfApi, operations: list[CommitOperation], message: str) -> None:
            """Split the commit into multiple parts if necessary.

            The HuggingFace API may time out if there are too many operations in a single commit.
            """
            import math

            max_ops = self.config.max_operations_per_commit
            num_commits = math.ceil(len(operations) / max_ops)
            for i in range(num_commits):
                begin = i * max_ops
                end = (i + 1) * max_ops
                part = operations[begin:end]
                commit_message = message + (f" (part {i:05d}-of-{num_commits:05d})" if num_commits > 1 else "")
                api.create_commit(
                    repo_id=self.repo,
                    repo_type="dataset",
                    revision=self.revision,
                    operations=part,
                    commit_message=commit_message,
                    commit_description=self.commit_description,
                )

            for op in operations:
                if isinstance(op, CommitOperationAdd):
                    output["path_in_repo"].append(op.path_in_repo)
                    output["operation"].append("ADD")
                    output["src_path_in_repo"].append("")
                elif isinstance(op, CommitOperationCopy):
                    output["path_in_repo"].append(op.path_in_repo)
                    output["operation"].append("COPY")
                    output["src_path_in_repo"].append(op.src_path_in_repo)
                elif isinstance(op, CommitOperationDelete):
                    output["path_in_repo"].append(op.path_in_repo)
                    output["operation"].append("DELETE")
                    output["src_path_in_repo"].append("")
                else:
                    raise ValueError(f"Expected `operations` to be a list of `CommitOperation` type, found: {type(op)}")

        def format_path(i: int) -> str:
            return f"{self.path_prefix}-{i:05d}-of-{count_new + count_existing:05d}.parquet"

        def rename(old_path: str, new_path: str) -> Iterator[CommitOperation]:
            if old_path != new_path:
                yield CommitOperationCopy(src_path_in_repo=old_path, path_in_repo=new_path)
                yield CommitOperationDelete(path_in_repo=old_path)

        with self._hf_api() as api:
            # In overwrite mode, delete existing files
            if self.overwrite:
                for obj in list_split(api):
                    # Delete old file
                    operations[obj.path] = CommitOperationDelete(
                        path_in_repo=obj.path, is_folder=isinstance(obj, RepoFolder)
                    )
            # In append mode, rename existing files to have the correct total number of parts
            else:
                rename_operations: list[CommitOperation] = []
                existing = list(obj for obj in list_split(api) if isinstance(obj, RepoFile))
                count_existing = len(existing)
                for i, obj in enumerate(existing):
                    new_path = format_path(i)
                    rename_operations.extend(rename(obj.path, new_path))
                # Rename files in a separate commit to prevent them from being overwritten by new files of the same name
                create_commits(
                    api,
                    operations=rename_operations,
                    message=f"{self.commit_message} (rename existing files before uploading new files)",
                )

            # Rename additions, putting them after existing files if any
            for i, addition in enumerate(additions):
                addition.path_in_repo = format_path(i + count_existing)
                # Overwrite the deletion operation if the file already exists
                operations[addition.path_in_repo] = addition

            # Upload the new files
            create_commits(api, operations=list(operations.values()), message=self.commit_message)

        return MicroPartition.from_pydict(output)
